# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.base.exchange import Exchange
from ccxt.abstract.hibachi import ImplicitAPI
import hashlib
from ccxt.base.types import Any, Balances, Currencies, Currency, DepositAddress, Int, LedgerEntry, Market, Num, Order, OrderBook, OrderRequest, OrderSide, OrderType, Position, Str, Strings, Ticker, FundingRate, Trade, TradingFees, Transaction
from typing import List
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import BadRequest
from ccxt.base.errors import OrderNotFound
from ccxt.base.decimal_to_precision import TICK_SIZE
from ccxt.base.precise import Precise


class hibachi(Exchange, ImplicitAPI):

    def describe(self) -> Any:
        return self.deep_extend(super(hibachi, self).describe(), {
            'id': 'hibachi',
            'name': 'Hibachi',
            'countries': ['US'],
            'rateLimit': 100,
            'userAgent': self.userAgents['chrome'],
            'certified': False,
            'pro': False,
            'dex': True,
            'has': {
                'CORS': None,
                'spot': False,
                'margin': False,
                'swap': True,
                'future': False,
                'option': False,
                'addMargin': False,
                'cancelAllOrders': True,
                'cancelOrder': True,
                'cancelOrders': True,
                'cancelWithdraw': False,
                'closeAllPositions': False,
                'closePosition': False,
                'createConvertTrade': False,
                'createDepositAddress': False,
                'createMarketBuyOrderWithCost': False,
                'createMarketOrder': False,
                'createMarketOrderWithCost': False,
                'createMarketSellOrderWithCost': False,
                'createOrder': True,
                'createOrders': True,
                'createOrderWithTakeProfitAndStopLoss': False,
                'createReduceOnlyOrder': False,
                'createStopLimitOrder': False,
                'createStopLossOrder': False,
                'createStopMarketOrder': False,
                'createStopOrder': False,
                'createTakeProfitOrder': False,
                'createTrailingAmountOrder': False,
                'createTrailingPercentOrder': False,
                'createTriggerOrder': False,
                'editOrder': True,
                'editOrders': True,
                'fetchAccounts': False,
                'fetchBalance': True,
                'fetchCanceledOrders': False,
                'fetchClosedOrder': False,
                'fetchClosedOrders': False,
                'fetchConvertCurrencies': False,
                'fetchConvertQuote': False,
                'fetchCurrencies': True,
                'fetchDepositAddress': True,
                'fetchDeposits': True,
                'fetchDepositsWithdrawals': False,
                'fetchFundingHistory': False,
                'fetchFundingInterval': False,
                'fetchFundingIntervals': False,
                'fetchFundingRate': True,
                'fetchFundingRateHistory': True,
                'fetchFundingRates': False,
                'fetchIndexOHLCV': False,
                'fetchLedger': True,
                'fetchLeverage': False,
                'fetchMarginAdjustmentHistory': False,
                'fetchMarginMode': False,
                'fetchMarkets': True,
                'fetchMyTrades': True,
                'fetchOHLCV': True,
                'fetchOpenInterest': True,
                'fetchOpenInterestHistory': False,
                'fetchOpenOrder': False,
                'fetchOpenOrders': True,
                'fetchOrder': True,
                'fetchOrderBook': True,
                'fetchOrders': False,
                'fetchOrderTrades': False,
                'fetchPosition': False,
                'fetchPositionMode': False,
                'fetchPositions': True,
                'fetchPremiumIndexOHLCV': False,
                'fetchStatus': False,
                'fetchTicker': True,
                'fetchTickers': False,
                'fetchTime': True,
                'fetchTrades': True,
                'fetchTradingFee': False,
                'fetchTradingFees': True,
                'fetchTradingLimits': False,
                'fetchTransactions': 'emulated',
                'fetchTransfers': False,
                'fetchWithdrawals': True,
                'reduceMargin': False,
                'setLeverage': False,
                'setMargin': False,
                'setPositionMode': False,
                'transfer': False,
                'withdraw': True,
            },
            'timeframes': {
                '1m': '1min',
                '5m': '5min',
                '15m': '15min',
                '1h': '1h',
                '4h': '4h',
                '1d': '1d',
                '1w': '1w',
            },
            'urls': {
                'logo': 'https://github.com/user-attachments/assets/7301bbb1-4f27-4167-8a55-75f74b14e973',
                'api': {
                    'public': 'https://data-api.hibachi.xyz',
                    'private': 'https://api.hibachi.xyz',
                },
                'www': 'https://www.hibachi.xyz/',
                'referral': {
                    'url': 'hibachi.xyz/r/ZBL2YFWIHU',
                },
            },
            'api': {
                'public': {
                    'get': {
                        'market/exchange-info': 1,
                        'market/data/trades': 1,
                        'market/data/prices': 1,
                        'market/data/stats': 1,
                        'market/data/klines': 1,
                        'market/data/orderbook': 1,
                        'market/data/open-interest': 1,
                        'market/data/funding-rates': 1,
                        'exchange/utc-timestamp': 1,
                    },
                },
                'private': {
                    'get': {
                        'capital/deposit-info': 1,
                        'capital/history': 1,
                        'trade/account/trading_history': 1,
                        'trade/account/info': 1,
                        'trade/order': 1,
                        'trade/account/trades': 1,
                        'trade/orders': 1,
                    },
                    'put': {
                        'trade/order': 1,
                    },
                    'delete': {
                        'trade/order': 1,
                        'trade/orders': 1,
                    },
                    'post': {
                        'trade/order': 1,
                        'trade/orders': 1,
                        'capital/withdraw': 1,
                    },
                },
            },
            'requiredCredentials': {
                'apiKey': True,
                'secret': False,
                'accountId': True,
                'privateKey': True,
            },
            'fees': {
                'trading': {
                    'tierBased': False,
                    'percentage': True,
                    'maker': self.parse_number('0.00015'),
                    'taker': self.parse_number('0.00045'),
                },
            },
            'options': {
            },
            'features': {
                'default': {
                    'sandbox': False,
                    'createOrder': {
                        'marginMode': False,
                        'triggerPrice': True,
                        'triggerPriceType': None,
                        'triggerDirection': None,
                        'stopLossPrice': False,
                        'takeProfitPrice': False,
                        'attachedStopLossTakeProfit': None,
                        'timeInForce': {
                            'IOC': True,
                            'FOK': False,
                            'PO': True,
                            'GTD': False,
                        },
                        'hedged': False,
                        'trailing': False,
                        'leverage': False,
                        'marketBuyByCost': False,
                        'marketBuyRequiresPrice': False,
                        'selfTradePrevention': False,
                        'iceberg': False,
                    },
                    'createOrders': None,
                    'fetchMyTrades': {
                        'marginMode': False,
                        'limit': None,
                        'daysBack': None,
                        'untilDays': None,
                        'symbolRequired': False,
                    },
                    'fetchOrder': {
                        'marginMode': False,
                        'trigger': False,
                        'trailing': False,
                        'symbolRequired': False,
                    },
                    'fetchOpenOrders': {
                        'marginMode': False,
                        'limit': None,
                        'trigger': False,
                        'trailing': False,
                        'symbolRequired': False,
                    },
                    'fetchOrders': None,
                    'fetchClosedOrders': None,
                    'fetchOHLCV': {
                        'limit': None,
                    },
                },
                'swap': {
                    'linear': {
                        'extends': 'default',
                    },
                    'inverse': None,
                },
                'future': {
                    'linear': {
                        'extends': 'default',
                    },
                    'inverse': None,
                },
            },
            'commonCurrencies': {},
            'exceptions': {
                'exact': {
                    '2': BadRequest,  # {"errorCode":2,"message":"Invalid signature: Failed to verify signature"}
                    '3': OrderNotFound,  # {"errorCode":3,"message":"Not found: order ID 33","status":"failed"}
                    '4': BadRequest,  # {"errorCode":4,"message":"Missing accountId","status":"failed"}
                },
                'broad': {
                },
            },
            'precisionMode': TICK_SIZE,
        })

    def get_account_id(self):
        self.check_required_credentials()
        id = self.parse_to_int(self.accountId)
        return id

    def parse_market(self, market: dict) -> Market:
        marketId = self.safe_string(market, 'symbol')
        numericId = self.safe_number(market, 'id')
        marketType = 'swap'
        baseId = self.safe_string(market, 'underlyingSymbol')
        quoteId = self.safe_string(market, 'settlementSymbol')
        base = self.safe_currency_code(baseId)
        quote = self.safe_currency_code(quoteId)
        settleId: Str = self.safe_string(market, 'settlementSymbol')
        settle: Str = self.safe_currency_code(settleId)
        symbol = base + '/' + quote + ':' + settle
        created = self.safe_integer_product(market, 'marketCreationTimestamp', 1000)
        return {
            'id': marketId,
            'numericId': numericId,
            'symbol': symbol,
            'base': base,
            'quote': quote,
            'settle': settle,
            'baseId': baseId,
            'quoteId': quoteId,
            'settleId': settleId,
            'type': marketType,
            'spot': False,
            'margin': False,
            'swap': True,
            'future': False,
            'option': False,
            'active': self.safe_string(market, 'status') == 'LIVE',
            'contract': True,
            'linear': True,
            'inverse': False,
            'contractSize': self.parse_number('1'),
            'expiry': None,
            'expiryDatetime': None,
            'strike': None,
            'optionType': None,
            'precision': {
                'amount': self.parse_number(self.parse_precision(self.safe_string(market, 'underlyingDecimals'))),
                'price': self.parse_number(self.safe_list(market, 'orderbookGranularities')[0]) / 10000.0,
            },
            'limits': {
                'leverage': {
                    'min': None,
                    'max': None,
                },
                'amount': {
                    'min': None,
                    'max': None,
                },
                'price': {
                    'min': None,
                    'max': None,
                },
                'cost': {
                    'min': self.safe_number(market, 'minNotional'),
                    'max': None,
                },
            },
            'created': created,
            'info': market,
        }

    def fetch_markets(self, params={}) -> List[Market]:
        """
        retrieves data on all markets for hibachi

        https://api-doc.hibachi.xyz/#183981da-8df5-40a0-a155-da15015dd536

        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: an array of objects representing market data
        """
        response = self.publicGetMarketExchangeInfo(params)
        # {
        #     "displayName": "ETH/USDT Perps",
        #     "id": 1,
        #     "maintenanceFactorForPositions": "0.030000",
        #     "marketCloseTimestamp": null,
        #     "marketOpenTimestamp": null,
        #     "minNotional": "1",
        #     "minOrderSize": "0.000000001",
        #     "orderbookGranularities": [
        #         "0.01",
        #         "0.1",
        #         "1",
        #         "10"
        #     ],
        #     "riskFactorForOrders": "0.066667",
        #     "riskFactorForPositions": "0.030000",
        #     "settlementDecimals": 6,
        #     "settlementSymbol": "USDT",
        #     "status": "LIVE",
        #     "stepSize": "0.000000001",
        #     "symbol": "ETH/USDT-P",
        #     "tickSize": "0.000001",
        #     "underlyingDecimals": 9,
        #     "underlyingSymbol": "ETH"
        # },
        rows = self.safe_list(response, 'futureContracts')
        return self.parse_markets(rows)

    def fetch_currencies(self, params={}) -> Currencies:
        """
        fetches all available currencies on an exchange

        https://api-doc.hibachi.xyz/#183981da-8df5-40a0-a155-da15015dd536

        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an associative dictionary of currencies
        """
        # Hibachi only supports USDT on Arbitrum at self time
        # We don't have an API endpoint to expose self information yet
        result: dict = {}
        networks: dict = {}
        networkId = 'ARBITRUM'
        networks[networkId] = {
            'id': networkId,
            'network': networkId,
            'limits': {
                'withdraw': {
                    'min': None,
                    'max': None,
                },
                'deposit': {
                    'min': None,
                    'max': None,
                },
            },
            'active': None,
            'deposit': None,
            'withdraw': None,
            'info': {},
        }
        code = self.safe_currency_code('USDT')
        result[code] = self.safe_currency_structure({
            'id': 'USDT',
            'name': 'USDT',
            'type': 'fiat',
            'code': code,
            'precision': self.parse_number('0.000001'),
            'active': True,
            'fee': None,
            'networks': networks,
            'deposit': True,
            'withdraw': True,
            'limits': {
                'deposit': {
                    'min': None,
                    'max': None,
                },
                'withdraw': {
                    'min': None,
                    'max': None,
                },
            },
            'info': {},
        })
        return result

    def parse_balance(self, response) -> Balances:
        result: dict = {
            'info': response,
        }
        # Hibachi only supports USDT on Arbitrum at self time
        code = self.safe_currency_code('USDT')
        account = self.account()
        account['total'] = self.safe_string(response, 'balance')
        account['free'] = self.safe_string(response, 'maximalWithdraw')
        result[code] = account
        return self.safe_balance(result)

    def fetch_balance(self, params={}) -> Balances:
        """
        query for balance and get the amount of funds available for trading or funds locked in orders

        https://api-doc.hibachi.xyz/#69aafedb-8274-4e21-bbaf-91dace8b8f31

        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        request: dict = {
            'accountId': self.get_account_id(),
        }
        response = self.privateGetTradeAccountInfo(self.extend(request, params))
        #
        # {
        #     assets: [{quantity: '3.000000', symbol: 'USDT'}],
        #     balance: '3.000000',
        #     maximalWithdraw: '3.000000',
        #     numFreeTransfersRemaining: '100',
        #     positions: [],
        #     totalOrderNotional: '0.000000',
        #     totalPositionNotional: '0.000000',
        #     totalUnrealizedFundingPnl: '0.000000',
        #     totalUnrealizedPnl: '0.000000',
        #     totalUnrealizedTradingPnl: '0.000000',
        #     tradeMakerFeeRate: '0.00000000',
        #     tradeTakerFeeRate: '0.00020000'
        # }
        #
        return self.parse_balance(response)

    def parse_ticker(self, ticker: dict, market: Market = None) -> Ticker:
        prices = self.safe_dict(ticker, 'prices')
        stats = self.safe_dict(ticker, 'stats')
        bid = self.safe_number(prices, 'bidPrice')
        ask = self.safe_number(prices, 'askPrice')
        last = self.safe_number(prices, 'tradePrice')
        high = self.safe_number(stats, 'high24h')
        low = self.safe_number(stats, 'low24h')
        volume = self.safe_number(stats, 'volume24h')
        return self.safe_ticker({
            'symbol': self.safe_symbol(None, market),
            'timestamp': None,
            'datetime': None,
            'bid': bid,
            'ask': ask,
            'last': last,
            'high': high,
            'low': low,
            'bidVolume': None,
            'askVolume': None,
            'vwap': None,
            'open': None,
            'close': last,
            'previousClose': None,
            'change': None,
            'percentage': None,
            'average': None,
            'baseVolume': None,
            'quoteVolume': volume,
            'info': ticker,
        }, market)

    def parse_trade(self, trade: dict, market: Market = None) -> Trade:
        # public fetchTrades:
        #      {
        #          "price": "3512.431902",
        #          "quantity": "1.414780098",
        #          "takerSide": "Buy",
        #          "timestamp": 1712692147
        #      }
        #
        # private fetchMyTrades:
        #      {
        #          "askAccountId": 221,
        #          "askOrderId": 589168494921909200,
        #          "bidAccountId": 132,
        #          "bidOrderId": 589168494829895700,
        #          "fee": "0.000477",
        #          "id": 199511136,
        #          "orderType": "MARKET",
        #          "price": "119257.90000",
        #          "quantity": "0.0000200000",
        #          "realizedPnl": "-0.000352",
        #          "side": "Sell",
        #          "symbol": "BTC/USDT-P",
        #          "timestamp": 1752543391
        #      }
        marketId = self.safe_string(trade, 'symbol')
        market = self.safe_market(marketId, market)
        symbol = market['symbol']
        id = self.safe_string(trade, 'id')
        price = self.safe_string(trade, 'price')
        amount = self.safe_string(trade, 'quantity')
        timestamp = self.safe_integer_product(trade, 'timestamp', 1000)
        cost = Precise.string_mul(price, amount)
        side = None
        fee = None
        orderType = None
        orderId = None
        takerOrMaker = None
        if id is None:
            # public trades
            side = self.safe_string_lower(trade, 'takerSide')
            takerOrMaker = 'taker'
        else:
            # private trades
            side = self.safe_string_lower(trade, 'side')
            fee = {'cost': self.safe_string(trade, 'fee'), 'currency': 'USDT'}
            orderType = self.safe_string_lower(trade, 'orderType')
            if side == 'buy':
                orderId = self.safe_string(trade, 'bidOrderId')
            else:
                orderId = self.safe_string(trade, 'askOrderId')
        return self.safe_trade({
            'id': id,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': symbol,
            'side': side,
            'price': price,
            'amount': amount,
            'cost': cost,
            'order': orderId,
            'takerOrMaker': takerOrMaker,
            'type': orderType,
            'fee': fee,
            'info': trade,
        }, market)

    def fetch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a particular symbol

        https://api-doc.hibachi.xyz/#86a53bc1-d3bb-4b93-8a11-7034d4698caa

        :param str symbol: unified market symbol
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch(maximum value is 100)
        :param dict [params]: extra parameters specific to the hibachi api endpoint
        :returns dict[]: a list of recent [trade structures]
        """
        self.load_markets()
        market = self.market(symbol)
        request = {
            'symbol': market['id'],
        }
        response = self.publicGetMarketDataTrades(self.extend(request, params))
        #
        # {
        #     "trades": [
        #         {
        #             "price": "111091.38352",
        #             "quantity": "0.0090090093",
        #             "takerSide": "Buy",
        #             "timestamp": 1752095479
        #         },
        #     ]
        # }
        #
        trades = self.safe_list(response, 'trades', [])
        return self.parse_trades(trades, market)

    def fetch_ticker(self, symbol: Str, params={}) -> Ticker:
        """

        https://api-doc.hibachi.xyz/#4abb30c4-e5c7-4b0f-9ade-790111dbfa47

        fetches a price ticker and the related information for the past 24h
        :param str symbol: unified symbol of the market
        :param dict [params]: extra parameters specific to the hibachi api endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request: dict = {
            'symbol': market['id'],
        }
        rawPromises = [
            self.publicGetMarketDataPrices(self.extend(request, params)),
            self.publicGetMarketDataStats(self.extend(request, params)),
        ]
        promises = rawPromises
        pricesResponse = promises[0]
        # {
        #     "askPrice": "3514.650296",
        #     "bidPrice": "3513.596112",
        #     "fundingRateEstimation": {
        #         "estimatedFundingRate": "0.000001",
        #         "nextFundingTimestamp": 1712707200
        #     },
        #     "markPrice": "3514.288858",
        #     "spotPrice": "3514.715000",
        #     "symbol": "ETH/USDT-P",
        #     "tradePrice": "2372.746570"
        # }
        statsResponse = promises[1]
        # {
        #     "high24h": "3819.507827",
        #     "low24h": "3754.474162",
        #     "symbol": "ETH/USDT-P",
        #     "volume24h": "23554.858590416"
        # }
        ticker = {
            'prices': pricesResponse,
            'stats': statsResponse,
        }
        return self.parse_ticker(ticker, market)

    def parse_order_status(self, status: str) -> str:
        statuses: dict = {
            'PENDING': 'open',
            'CHILD_PENDING': 'open',
            'SCHEDULED_TWAP': 'open',
            'PLACED': 'open',
            'PARTIALLY_FILLED': 'open',
            'FILLED': 'closed',
            'CANCELLED': 'canceled',
            'REJECTED': 'rejected',
        }
        return self.safe_string(statuses, status, status)

    def parse_order(self, order: dict, market: Market = None) -> Order:
        marketId = self.safe_string(order, 'symbol')
        market = self.safe_market(marketId, market)
        status = self.safe_string(order, 'status')
        type = self.safe_string_lower(order, 'orderType')
        price = self.safe_string(order, 'price')
        rawSide = self.safe_string(order, 'side')
        side = None
        if rawSide == 'BID':
            side = 'buy'
        elif rawSide == 'ASK':
            side = 'sell'
        amount = self.safe_string(order, 'totalQuantity')
        remaining = self.safe_string(order, 'availableQuantity')
        totalQuantity = self.safe_string(order, 'totalQuantity')
        availableQuantity = self.safe_string(order, 'availableQuantity')
        filled = None
        if totalQuantity is not None and availableQuantity is not None:
            filled = Precise.string_sub(totalQuantity, availableQuantity)
        timeInForce = 'GTC'
        orderFlags = self.safe_value(order, 'orderFlags')
        postOnly = False
        reduceOnly = False
        if orderFlags == 'POST_ONLY':
            timeInForce = 'PO'
            postOnly = True
        elif orderFlags == 'IOC':
            timeInForce = 'IOC'
        elif orderFlags == 'REDUCE_ONLY':
            reduceOnly = True
        return self.safe_order({
            'info': order,
            'id': self.safe_string(order, 'orderId'),
            'clientOrderId': None,
            'datetime': None,
            'timestamp': None,
            'lastTradeTimestamp': None,
            'lastUpdateTimestamp': None,
            'status': self.parse_order_status(status),
            'symbol': market['symbol'],
            'type': type,
            'timeInForce': timeInForce,
            'side': side,
            'price': price,
            'average': None,
            'amount': amount,
            'filled': filled,
            'remaining': remaining,
            'cost': None,
            'trades': None,
            'fee': None,
            'reduceOnly': reduceOnly,
            'postOnly': postOnly,
            'triggerPrice': self.safe_number(order, 'triggerPrice'),
        }, market)

    def fetch_order(self, id: str, symbol: Str = None, params={}) -> Order:
        """
        fetches information on an order made by the user

        https://api-doc.hibachi.xyz/#096a8854-b918-4de8-8731-b2a28d26b96d

        :param str id: the order id
        :param str symbol: unified symbol of the market the order was made in
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        market = None
        if symbol is not None:
            market = self.market(symbol)
        request: dict = {
            'orderId': id,
            'accountId': self.get_account_id(),
        }
        response = self.privateGetTradeOrder(self.extend(request, params))
        return self.parse_order(response, market)

    def fetch_trading_fees(self, params={}) -> TradingFees:
        """
        fetch the trading fee
 @param params extra parameters
        :returns dict: a map of market symbols to `fee structures <https://docs.ccxt.com/#/?id=fee-structure>`
        """
        self.load_markets()
        request: dict = {
            'accountId': self.get_account_id(),
        }
        response = self.privateGetTradeAccountInfo(self.extend(request, params))
        #    {
        #        "tradeMakerFeeRate": "0.00000000",
        #        "tradeTakerFeeRate": "0.00020000"
        #    },
        makerFeeRate = self.safe_number(response, 'tradeMakerFeeRate')
        takerFeeRate = self.safe_number(response, 'tradeTakerFeeRate')
        result: dict = {}
        for i in range(0, len(self.symbols)):
            symbol = self.symbols[i]
            result[symbol] = {
                'info': response,
                'symbol': symbol,
                'maker': makerFeeRate,
                'taker': takerFeeRate,
                'percentage': True,
            }
        return result

    def order_message(self, market, nonce: float, feeRate: float, type: OrderType, side: OrderSide, amount: float, price: Num = None):
        sideInternal = 0
        if side == 'sell':
            sideInternal = 0
        elif side == 'buy':
            sideInternal = 1
        # Converting them to internal representation:
        # - Quantity: Internal = External * (10^underlyingDecimals)
        # - Price: Internal = External * (2^32) * (10^(settlementDecimals-underlyingDecimals))
        # - FeeRate: Internal = External * (10^8)
        amountStr = self.amount_to_precision(self.safe_string(market, 'symbol'), amount)
        feeRateStr = self.number_to_string(feeRate)
        info = self.safe_dict(market, 'info')
        underlying = '1e' + self.safe_string(info, 'underlyingDecimals')
        settlement = '1e' + self.safe_string(info, 'settlementDecimals')
        one = '1'
        feeRateFactor = '100000000'  # 10^8
        priceFactor = '4294967296'  # 2^32
        quantityInternal = Precise.string_div(Precise.string_mul(amountStr, underlying), one, 0)
        feeRateInternal = Precise.string_div(Precise.string_mul(feeRateStr, feeRateFactor), one, 0)
        # Encoding
        nonce16 = self.int_to_base16(nonce)
        noncePadded = nonce16.rjust(16, '0')
        encodedNonce = self.base16_to_binary(noncePadded)
        numericId = self.int_to_base16(self.safe_integer(market, 'numericId'))
        numericIdPadded = numericId.rjust(8, '0')
        encodedMarketId = self.base16_to_binary(numericIdPadded)
        quantity16 = self.int_to_base16(self.parse_to_int(quantityInternal))
        quantityPadded = quantity16.rjust(16, '0')
        encodedQuantity = self.base16_to_binary(quantityPadded)
        sideInternal16 = self.int_to_base16(sideInternal)
        sidePadded = sideInternal16.rjust(8, '0')
        encodedSide = self.base16_to_binary(sidePadded)
        feeRateInternal16 = self.int_to_base16(self.parse_to_int(feeRateInternal))
        feeRatePadded = feeRateInternal16.rjust(16, '0')
        encodedFeeRate = self.base16_to_binary(feeRatePadded)
        encodedPrice = self.binary_concat()
        if type == 'limit':
            priceStr = self.price_to_precision(self.safe_string(market, 'symbol'), price)
            priceInternal = Precise.string_div(Precise.string_div(Precise.string_mul(Precise.string_mul(priceStr, priceFactor), settlement), underlying), one, 0)
            price16 = self.int_to_base16(self.parse_to_int(priceInternal))
            pricePadded = price16.rjust(16, '0')
            encodedPrice = self.base16_to_binary(pricePadded)
        message = self.binary_concat(encodedNonce, encodedMarketId, encodedQuantity, encodedSide, encodedPrice, encodedFeeRate)
        return message

    def create_order_request(self, nonce: float, symbol: str, type: OrderType, side: OrderSide, amount: float, price: Num = None, params={}):
        market = self.market(symbol)
        feeRate = max(self.safe_number(market, 'taker'), self.safe_number(market, 'maker'))
        sideInternal = ''
        if side == 'sell':
            sideInternal = 'ASK'
        elif side == 'buy':
            sideInternal = 'BID'
        priceInternal = ''
        if price:
            priceInternal = self.price_to_precision(symbol, price)
        message = self.order_message(market, nonce, feeRate, type, side, amount, price)
        signature = self.sign_message(message, self.privateKey)
        request = {
            'symbol': self.safe_string(market, 'id'),
            'nonce': nonce,
            'side': sideInternal,
            'orderType': type.upper(),
            'quantity': self.amount_to_precision(symbol, amount),
            'price': priceInternal,
            'signature': signature,
            'maxFeesPercent': self.number_to_string(feeRate),
        }
        postOnly = self.is_post_only(type.upper() == 'MARKET', None, params)
        reduceOnly = self.safe_bool_2(params, 'reduceOnly', 'reduce_only')
        timeInForce = self.safe_string_lower(params, 'timeInForce')
        triggerPrice = self.safe_string_2(params, 'triggerPrice', 'stopPrice')
        if postOnly:
            request['orderFlags'] = 'POST_ONLY'
        elif timeInForce == 'ioc':
            request['orderFlags'] = 'IOC'
        elif reduceOnly:
            request['orderFlags'] = 'REDUCE_ONLY'
        if triggerPrice is not None:
            request['triggerPrice'] = triggerPrice
        params = self.omit(params, ['reduceOnly', 'reduce_only', 'postOnly', 'timeInForce', 'stopPrice', 'triggerPrice'])
        return self.extend(request, params)

    def create_order(self, symbol: str, type: OrderType, side: OrderSide, amount: float, price: Num = None, params={}):
        """
        create a trade order

        https://api-doc.hibachi.xyz/#00f6d5ad-5275-41cb-a1a8-19ed5d142124

        :param str symbol: unified symbol of the market to create an order in
        :param str type: 'market' or 'limit'
        :param str side: 'buy' or 'sell'
        :param float amount: how much of currency you want to trade in units of base currency
        :param float [price]: the price at which the order is to be fulfilled, in units of the quote currency, ignored in market orders
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        nonce = self.nonce()
        request = self.create_order_request(nonce, symbol, type, side, amount, price, params)
        request['accountId'] = self.get_account_id()
        response = self.privatePostTradeOrder(request)
        #
        # {
        #     "orderId": "578721673790138368"
        # }
        #
        return self.safe_order({
            'id': self.safe_string(response, 'orderId'),
            'status': 'pending',
        })

    def create_orders(self, orders: List[OrderRequest], params={}) -> List[Order]:
        """
        *contract only* create a list of trade orders

        https://api-doc.hibachi.xyz/#c2840b9b-f02c-44ed-937d-dc2819f135b4

        :param Array orders: list of orders to create, each object should contain the parameters required by createOrder, namely symbol, type, side, amount, price and params
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        nonce = self.nonce()
        requestOrders = []
        for i in range(0, len(orders)):
            rawOrder = orders[i]
            symbol = self.safe_string(rawOrder, 'symbol')
            type = self.safe_string(rawOrder, 'type')
            side = self.safe_string(rawOrder, 'side')
            amount = self.safe_value(rawOrder, 'amount')
            price = self.safe_value(rawOrder, 'price')
            orderParams = self.safe_dict(rawOrder, 'params', {})
            orderRequest = self.create_order_request(nonce + i, symbol, type, side, amount, price, orderParams)
            orderRequest['action'] = 'place'
            requestOrders.append(orderRequest)
        request: dict = {
            'accountId': self.get_account_id(),
            'orders': requestOrders,
        }
        response = self.privatePostTradeOrders(self.extend(request, params))
        #
        # {"orders": [{nonce: '1754349993908', orderId: '589642085255349248'}]}
        #
        ret = []
        responseOrders = self.safe_list(response, 'orders')
        for i in range(0, len(responseOrders)):
            responseOrder = responseOrders[i]
            ret.append(self.safe_order({
                'info': responseOrder,
                'id': self.safe_string(responseOrder, 'orderId'),
                'status': 'pending',
            }))
        return ret

    def edit_order_request(self, nonce: float, id: str, symbol: str, type: OrderType, side: OrderSide, amount: Num = None, price: Num = None, params={}):
        market = self.market(symbol)
        feeRate = max(self.safe_number(market, 'taker'), self.safe_number(market, 'maker'))
        message = self.order_message(market, nonce, feeRate, type, side, amount, price)
        signature = self.sign_message(message, self.privateKey)
        request = {
            'orderId': id,
            'nonce': nonce,
            'updatedQuantity': self.amount_to_precision(symbol, amount),
            'updatedPrice': self.price_to_precision(symbol, price),
            'maxFeesPercent': self.number_to_string(feeRate),
            'signature': signature,
        }
        return self.extend(request, params)

    def edit_order(self, id: str, symbol: str, type: OrderType, side: OrderSide, amount: Num = None, price: Num = None, params={}):
        """
        edit a limit order that is not matched

        https://api-doc.hibachi.xyz/#94d2cdaf-1c71-440f-a981-da1112824810

        :param str id: order id
        :param str symbol: unified symbol of the market to create an order in
        :param str type: must be 'limit'
        :param str side: 'buy' or 'sell', should stay the same with original side
        :param float amount: how much of currency you want to trade in units of base currency
        :param float [price]: the price at which the order is to be fulfilled, in units of the quote currency
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        nonce = self.nonce()
        request = self.edit_order_request(nonce, id, symbol, type, side, amount, price, params)
        request['accountId'] = self.get_account_id()
        self.privatePutTradeOrder(request)
        # At self time the response body is empty. A 200 response means the update request is accepted and sent to process
        #
        # {}
        #
        return self.safe_order({
            'id': id,
            'status': 'pending',
        })

    def edit_orders(self, orders: List[OrderRequest], params={}) -> List[Order]:
        """
        edit a list of trade orders

        https://api-doc.hibachi.xyz/#c2840b9b-f02c-44ed-937d-dc2819f135b4

        :param Array orders: list of orders to edit, each object should contain the parameters required by editOrder, namely id, symbol, type, side, amount, price and params
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        nonce = self.nonce()
        requestOrders = []
        for i in range(0, len(orders)):
            rawOrder = orders[i]
            id = self.safe_string(rawOrder, 'id')
            symbol = self.safe_string(rawOrder, 'symbol')
            type = self.safe_string(rawOrder, 'type')
            side = self.safe_string(rawOrder, 'side')
            amount = self.safe_value(rawOrder, 'amount')
            price = self.safe_value(rawOrder, 'price')
            orderParams = self.safe_dict(rawOrder, 'params', {})
            orderRequest = self.edit_order_request(nonce + i, id, symbol, type, side, amount, price, orderParams)
            orderRequest['action'] = 'modify'
            requestOrders.append(orderRequest)
        request: dict = {
            'accountId': self.get_account_id(),
            'orders': requestOrders,
        }
        response = self.privatePostTradeOrders(self.extend(request, params))
        #
        # {"orders": [{"orderId": "589636801329628160"}]}
        #
        ret = []
        responseOrders = self.safe_list(response, 'orders')
        for i in range(0, len(responseOrders)):
            responseOrder = responseOrders[i]
            ret.append(self.safe_order({
                'info': responseOrder,
                'id': self.safe_string(responseOrder, 'orderId'),
                'status': 'pending',
            }))
        return ret

    def cancel_order_request(self, id: str):
        bigid = self.convert_to_big_int(id)
        idbase16 = self.int_to_base16(bigid)
        idPadded = idbase16.rjust(16, '0')
        message = self.base16_to_binary(idPadded)
        signature = self.sign_message(message, self.privateKey)
        return {
            'orderId': id,
            'signature': signature,
        }

    def cancel_order(self, id: str, symbol: Str = None, params={}):
        """

        https://api-doc.hibachi.xyz/#e99c4f48-e610-4b7c-b7f6-1b4bb7af0271

        cancels an open order
        :param str id: order id
        :param str symbol: is unused
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: An `order structure <https://docs.ccxt.com/#/?id=order-structure>`
        """
        request: dict = self.cancel_order_request(id)
        request['accountId'] = self.get_account_id()
        response = self.privateDeleteTradeOrder(self.extend(request, params))
        # At self time the response body is empty. A 200 response means the cancel request is accepted and sent to cancel
        #
        # {}
        #
        return self.safe_order({
            'info': response,
            'id': id,
            'status': 'canceled',
        })

    def cancel_orders(self, ids: List[str], symbol: Str = None, params={}):
        """
        cancel multiple orders

        https://api-doc.hibachi.xyz/#c2840b9b-f02c-44ed-937d-dc2819f135b4

        :param str[] ids: order ids
        :param str [symbol]: unified market symbol, unused
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        orders = []
        for i in range(0, len(ids)):
            orderRequest = self.cancel_order_request(ids[i])
            orderRequest['action'] = 'cancel'
            orders.append(orderRequest)
        request: dict = {
            'accountId': self.get_account_id(),
            'orders': orders,
        }
        response = self.privatePostTradeOrders(self.extend(request, params))
        #
        # {"orders": [{"orderId": "589636801329628160"}]}
        #
        ret = []
        responseOrders = self.safe_list(response, 'orders')
        for i in range(0, len(responseOrders)):
            responseOrder = responseOrders[i]
            ret.append(self.safe_order({
                'info': responseOrder,
                'id': self.safe_string(responseOrder, 'orderId'),
                'status': 'canceled',
            }))
        return ret

    def cancel_all_orders(self, symbol: Str = None, params={}):
        """

        https://api-doc.hibachi.xyz/#8ed24695-016e-49b2-a72d-7511ca921fee

        cancel all open orders in a market
        :param str symbol: unified market symbol
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        nonce = self.nonce()
        nonce16 = self.int_to_base16(nonce)
        noncePadded = nonce16.rjust(16, '0')
        message = self.base16_to_binary(noncePadded)
        signature = self.sign_message(message, self.privateKey)
        request: dict = {
            'accountId': self.get_account_id(),
            'nonce': nonce,
            'signature': signature,
        }
        if symbol is not None:
            market = self.market(symbol)
            request['contractId'] = self.safe_integer(market, 'numericId')
        response = self.privateDeleteTradeOrders(self.extend(request, params))
        # At self time the response body is empty. A 200 response means the cancel request is accepted and sent to process
        #
        # {}
        #
        return [
            self.safe_order({
                'info': response,
            }),
        ]

    def encode_withdraw_message(self, amount: float, maxFees: float, address: str):
        # Converting them to internal representation:
        # - Quantity: Internal = External * (10^6)
        # - maxFees: Internal = External * (10^6)
        # We only have USDT currency time
        USDTAssetId = 1
        USDTFactor = '1000000'
        amountStr = self.number_to_string(amount)
        maxFeesStr = self.number_to_string(maxFees)
        one = '1'
        quantityInternal = Precise.string_div(Precise.string_mul(amountStr, USDTFactor), one, 0)
        maxFeesInternal = Precise.string_div(Precise.string_mul(maxFeesStr, USDTFactor), one, 0)
        # Encoding
        usdtAsset16 = self.int_to_base16(USDTAssetId)
        usdtAssetPadded = usdtAsset16.rjust(8, '0')
        encodedAssetId = self.base16_to_binary(usdtAssetPadded)
        quantity16 = self.int_to_base16(self.parse_to_int(quantityInternal))
        quantityPadded = quantity16.rjust(16, '0')
        encodedQuantity = self.base16_to_binary(quantityPadded)
        maxFees16 = self.int_to_base16(self.parse_to_int(maxFeesInternal))
        maxFeesPadded = maxFees16.rjust(16, '0')
        encodedMaxFees = self.base16_to_binary(maxFeesPadded)
        encodedAddress = self.base16_to_binary(address)
        message = self.binary_concat(encodedAssetId, encodedQuantity, encodedMaxFees, encodedAddress)
        return message

    def withdraw(self, code: str, amount: float, address: str, tag: Str = None, params={}) -> Transaction:
        """
        make a withdrawal

        https://api-doc.hibachi.xyz/#6421625d-3e45-45fa-be9b-d2a0e780c090

        :param str code: unified currency code, only support USDT
        :param float amount: the amount to withdraw
        :param str address: the address to withdraw to
        :param str tag:
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `transaction structure <https://docs.ccxt.com/#/?id=transaction-structure>`
        """
        withdrawAddress = address[-40:]
        # Get the withdraw fees
        exchangeInfo = self.publicGetMarketExchangeInfo(params)
        # {
        #      "feeConfig": {
        #          "depositFees": "0.004518",
        #          "tradeMakerFeeRate": "0.00000000",
        #          "tradeTakerFeeRate": "0.00020000",
        #          "transferFeeRate": "0.00010000",
        #          "withdrawalFees": "0.012050"
        #    },
        # }
        feeConfig = self.safe_dict(exchangeInfo, 'feeConfig')
        maxFees = self.safe_number(feeConfig, 'withdrawalFees')
        # Generate the signature
        message = self.encode_withdraw_message(amount, maxFees, withdrawAddress)
        signature = self.sign_message(message, self.privateKey)
        request = {
            'accountId': self.get_account_id(),
            'coin': 'USDT',
            'network': 'ARBITRUM',
            'withdrawAddress': withdrawAddress,
            'selfWithdrawal': False,
            'quantity': self.number_to_string(amount),
            'maxFees': self.number_to_string(maxFees),
            'signature': signature,
        }
        self.privatePostCapitalWithdraw(self.extend(request, params))
        # At self time the response body is empty. A 200 response means the withdraw request is accepted and sent to process
        #
        # {}
        #
        return {
            'info': None,
            'id': None,
            'txid': None,
            'timestamp': self.milliseconds(),
            'datetime': None,
            'address': None,
            'addressFrom': None,
            'addressTo': withdrawAddress,
            'tag': None,
            'tagFrom': None,
            'tagTo': None,
            'type': 'withdrawal',
            'amount': amount,
            'currency': code,
            'status': 'pending',
            'fee': {'currency': 'USDT', 'cost': maxFees},
            'network': 'ARBITRUM',
            'updated': None,
            'comment': None,
            'internal': None,
        }

    def nonce(self):
        return self.milliseconds()

    def sign_message(self, message, privateKey):
        if len(privateKey) == 44:
            # For Exchange Managed account, the key length is 44 and we use HMAC to sign the message
            return self.hmac(message, self.encode(privateKey), hashlib.sha256, 'hex')
        else:
            # For Trustless account, the key length is 66 including '0x' and we use ECDSA to sign the message
            hash = self.hash(self.encode(message), 'sha256', 'hex')
            signature = self.ecdsa(hash[-64:], privateKey[-64:], 'secp256k1', None)
            r = signature['r']
            s = signature['s']
            v = signature['v']
            return r.rjust(64, '0') + s.rjust(64, '0') + self.int_to_base16(v).rjust(2, '0')

    def fetch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        fetches the state of the open orders on the orderbook

        https://api-doc.hibachi.xyz/#4abb30c4-e5c7-4b0f-9ade-790111dbfa47

        :param str symbol: unified symbol of the market
        :param int [limit]: currently unused
        :param dict [params]: extra parameters to be passed -- see documentation link above
        :returns dict: A dictionary containg `orderbook information <https://docs.ccxt.com/#/?id=order-book-structure>`
        """
        self.load_markets()
        market: Market = self.market(symbol)
        request: dict = {
            'symbol': market['id'],
        }
        response = self.publicGetMarketDataOrderbook(self.extend(request, params))
        formattedResponse = {}
        formattedResponse['ask'] = self.safe_list(self.safe_dict(response, 'ask'), 'levels')
        formattedResponse['bid'] = self.safe_list(self.safe_dict(response, 'bid'), 'levels')
        # {
        #     "ask": {
        #         "endPrice": "3512.63",
        #         "levels": [
        #             {
        #                 "price": "3511.93",
        #                 "quantity": "0.284772482"
        #             },
        #             {
        #                 "price": "3512.28",
        #                 "quantity": "0.569544964"
        #             },
        #             {
        #                 "price": "3512.63",
        #                 "quantity": "0.854317446"
        #             }
        #         ],
        #         "startPrice": "3511.93"
        #     },
        #     "bid": {
        #         "endPrice": "3510.87",
        #         "levels": [
        #             {
        #                 "price": "3515.39",
        #                 "quantity": "2.345153070"
        #             },
        #             {
        #                 "price": "3511.22",
        #                 "quantity": "0.284772482"
        #             },
        #             {
        #                 "price": "3510.87",
        #                 "quantity": "0.569544964"
        #             }
        #         ],
        #         "startPrice": "3515.39"
        #     }
        # }
        return self.parse_order_book(formattedResponse, symbol, self.milliseconds(), 'bid', 'ask', 'price', 'quantity')

    def fetch_my_trades(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}):
        """

        https://api-doc.hibachi.xyz/#0adbf143-189f-40e0-afdc-88af4cba3c79

        fetch all trades made by the user
        :param str symbol: unified market symbol
        :param int [since]: the earliest time in ms to fetch trades for
        :param int [limit]: the maximum number of trades structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns Trade[]: a list of `trade structures <https://docs.ccxt.com/#/?id=trade-structure>`
        """
        self.load_markets()
        market: Market = None
        if symbol is not None:
            market = self.market(symbol)
        request = {'accountId': self.get_account_id()}
        response = self.privateGetTradeAccountTrades(self.extend(request, params))
        #
        # {
        #     "trades": [
        #         {
        #             "askAccountId": 221,
        #             "askOrderId": 589168494921909200,
        #             "bidAccountId": 132,
        #             "bidOrderId": 589168494829895700,
        #             "fee": "0.000477",
        #             "id": 199511136,
        #             "orderType": "MARKET",
        #             "price": "119257.90000",
        #             "quantity": "0.0000200000",
        #             "realizedPnl": "-0.000352",
        #             "side": "Sell",
        #             "symbol": "BTC/USDT-P",
        #             "timestamp": 1752543391
        #         }
        #     ]
        # }
        #
        trades = self.safe_list(response, 'trades')
        return self.parse_trades(trades, market, since, limit, params)

    def parse_ohlcv(self, ohlcv, market: Market = None) -> list:
        #
        # [
        #     {
        #       "close": "3704.751036",
        #       "high": "3716.530378",
        #       "interval": "1h",
        #       "low": "3699.627883",
        #       "open": "3716.406894",
        #       "timestamp": 1712628000,
        #       "volumeNotional": "1637355.846362"
        #     }
        #   ]
        #
        return [
            self.safe_integer_product(ohlcv, 'timestamp', 1000),
            self.safe_number(ohlcv, 'open'),
            self.safe_number(ohlcv, 'high'),
            self.safe_number(ohlcv, 'low'),
            self.safe_number(ohlcv, 'close'),
            self.safe_number(ohlcv, 'volumeNotional'),
        ]

    def fetch_open_orders(self, symbol: str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        fetches all current open orders

        https://api-doc.hibachi.xyz/#3243f8a0-086c-44c5-ab8a-71bbb7bab403

        :param str [symbol]: unified market symbol to filter by
        :param int [since]: milisecond timestamp of the earliest order
        :param int [limit]: the maximum number of open orders to return
        :param dict [params]: extra parameters
        :returns Order[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        self.load_markets()
        market = None
        if symbol is not None:
            market = self.market(symbol)
        request = {
            'accountId': self.get_account_id(),
        }
        response = self.privateGetTradeOrders(self.extend(request, params))
        # [
        #     {
        #         "accountId": 12452,
        #         "availableQuantity": "0.0000230769",
        #         "contractId": 2,
        #         "creationTime": 1752684501,
        #         "orderId": "589205486123876352",
        #         "orderType": "LIMIT",
        #         "price": "130000.00000",
        #         "side": "ASK",
        #         "status": "PLACED",
        #         "symbol": "BTC/USDT-P",
        #         "totalQuantity": "0.0000230769"
        #     },
        #     {
        #         "accountId": 12452,
        #         "availableQuantity": "1.234000000",
        #         "contractId": 1,
        #         "creationTime": 1752240682,
        #         "orderId": "589089141754429441",
        #         "orderType": "LIMIT",
        #         "price": "1.234000",
        #         "side": "BID",
        #         "status": "PLACED",
        #         "symbol": "ETH/USDT-P",
        #         "totalQuantity": "1.234000000"
        #     }
        # ]
        return self.parse_orders(response, market, since, limit)

    def fetch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """

         https://api-doc.hibachi.xyz/#4f0eacec-c61e-4d51-afb3-23c51c2c6bac

        fetches historical candlestick data containing the close, high, low, open prices, interval and the volumeNotional
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param int [params.until]: timestamp in ms of the latest candle to fetch
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        self.load_markets()
        market = self.market(symbol)
        timeframe = self.safe_string(self.timeframes, timeframe, timeframe)
        request: dict = {
            'symbol': market['id'],
            'interval': timeframe,
        }
        if since is not None:
            request['fromMs'] = since
        until: Int = None
        until, params = self.handle_option_and_params(params, 'fetchOHLCV', 'until')
        if until is not None:
            request['toMs'] = until
        response = self.publicGetMarketDataKlines(self.extend(request, params))
        #
        # [
        #     {
        #       "close": "3704.751036",
        #       "high": "3716.530378",
        #       "interval": "1h",
        #       "low": "3699.627883",
        #       "open": "3716.406894",
        #       "timestamp": 1712628000,
        #       "volumeNotional": "1637355.846362"
        #     }
        #   ]
        #
        klines = self.safe_list(response, 'klines', [])
        return self.parse_ohlcvs(klines, market, timeframe, since, limit)

    def fetch_positions(self, symbols: Strings = None, params={}) -> List[Position]:
        """
        fetch all open positions

        https://api-doc.hibachi.xyz/#69aafedb-8274-4e21-bbaf-91dace8b8f31

        :param str[] [symbols]: list of unified market symbols
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `position structure <https://docs.ccxt.com/#/?id=position-structure>`
        """
        self.load_markets()
        symbols = self.market_symbols(symbols)
        request: dict = {
            'accountId': self.get_account_id(),
        }
        response = self.privateGetTradeAccountInfo(self.extend(request, params))
        #
        # {
        #     "assets": [
        #       {
        #         "quantity": "14.130626",
        #         "symbol": "USDT"
        #       }
        #     ],
        #     "balance": "14.186087",
        #     "maximalWithdraw": "4.152340",
        #     "numFreeTransfersRemaining": 96,
        #     "positions": [
        #       {
        #         "direction": "Short",
        #         "entryNotional": "10.302213",
        #         "notionalValue": "10.225008",
        #         "quantity": "0.004310550",
        #         "symbol": "ETH/USDT-P",
        #         "unrealizedFundingPnl": "0.000000",
        #         "unrealizedTradingPnl": "0.077204"
        #       },
        #       {
        #         "direction": "Short",
        #         "entryNotional": "2.000016",
        #         "notionalValue": "1.999390",
        #         "quantity": "0.0000328410",
        #         "symbol": "BTC/USDT-P",
        #         "unrealizedFundingPnl": "0.000000",
        #         "unrealizedTradingPnl": "0.000625"
        #       },
        #       {
        #         "direction": "Short",
        #         "entryNotional": "2.000015",
        #         "notionalValue": "2.022384",
        #         "quantity": "0.01470600",
        #         "symbol": "SOL/USDT-P",
        #         "unrealizedFundingPnl": "0.000000",
        #         "unrealizedTradingPnl": "-0.022369"
        #       }
        #     ],
        #   }
        #
        data = self.safe_list(response, 'positions', [])
        return self.parse_positions(data, symbols)

    def parse_position(self, position: dict, market: Market = None):
        #
        # {
        #     "direction": "Short",
        #     "entryNotional": "10.302213",
        #     "notionalValue": "10.225008",
        #     "quantity": "0.004310550",
        #     "symbol": "ETH/USDT-P",
        #     "unrealizedFundingPnl": "0.000000",
        #     "unrealizedTradingPnl": "0.077204"
        # }
        #
        marketId = self.safe_string(position, 'symbol')
        market = self.safe_market(marketId, market)
        symbol = market['symbol']
        side = self.safe_string_lower(position, 'direction')
        quantity = self.safe_string(position, 'quantity')
        unrealizedFunding = self.safe_string(position, 'unrealizedFundingPnl', '0')
        unrealizedTrading = self.safe_string(position, 'unrealizedTradingPnl', '0')
        unrealizedPnl = Precise.string_add(unrealizedFunding, unrealizedTrading)
        return self.safe_position({
            'info': position,
            'id': None,
            'symbol': symbol,
            'entryPrice': self.safe_string(position, 'average_entry_price'),
            'markPrice': None,
            'notional': self.safe_string(position, 'notionalValue'),
            'collateral': None,
            'unrealizedPnl': unrealizedPnl,
            'side': side,
            'contracts': self.parse_number(quantity),
            'contractSize': None,
            'timestamp': None,
            'datetime': None,
            'hedged': None,
            'maintenanceMargin': None,
            'maintenanceMarginPercentage': None,
            'initialMargin': None,
            'initialMarginPercentage': None,
            'leverage': None,
            'liquidationPrice': None,
            'marginRatio': None,
            'marginMode': None,
            'percentage': None,
        })

    def sign(self, path, api='public', method='GET', params={}, headers=None, body=None):
        endpoint = '/' + self.implode_params(path, params)
        url = self.urls['api'][api] + endpoint
        headers = {'Hibachi-Client': 'HibachiCCXT/unversioned'}
        if method == 'GET':
            request = self.omit(params, self.extract_params(path))
            query = self.urlencode(request)
            if len(query) != 0:
                url += '?' + query
        if method == 'POST' or method == 'PUT' or method == 'DELETE':
            headers['Content-Type'] = 'application/json'
            body = self.json(params)
        if api == 'private':
            self.check_required_credentials()
            headers['Authorization'] = self.apiKey
        return {'url': url, 'method': method, 'body': body, 'headers': headers}

    def handle_errors(self, httpCode: int, reason: str, url: str, method: str, headers: dict, body: str, response, requestHeaders, requestBody):
        if response is None:
            return None  # fallback to default error handler
        if 'status' in response:
            #
            #     {"errorCode":4,"message":"Invalid input: Invalid quantity: 0","status":"failed"}
            #
            status = self.safe_string(response, 'status')
            if status == 'failed':
                code = self.safe_string(response, 'errorCode')
                feedback = self.id + ' ' + body
                self.throw_broadly_matched_exception(self.exceptions['broad'], body, feedback)
                self.throw_exactly_matched_exception(self.exceptions['exact'], code, feedback)
                message = self.safe_string(response, 'message')
                self.throw_exactly_matched_exception(self.exceptions['exact'], message, feedback)
                raise ExchangeError(feedback)
        return None

    def parse_transaction_type(self, type):
        types: dict = {
            'deposit': 'transaction',
            'withdrawal': 'transaction',
            'transfer-in': 'transfer',
            'transfer-out': 'transfer',
        }
        return self.safe_string(types, type, type)

    def parse_transaction_status(self, status):
        statuses: dict = {
            'pending': 'pending',
            'claimable': 'pending',
            'completed': 'ok',
            'failed': 'canceled',
        }
        return self.safe_string(statuses, status, status)

    def parse_ledger_entry(self, item: dict, currency: Currency = None) -> LedgerEntry:
        transactionType = self.safe_string(item, 'transactionType')
        timestamp = None
        type = None
        direction = None
        amount = None
        fee = None
        referenceId = None
        referenceAccount = None
        status = None
        if transactionType is None:
            # response from TradeAccountTradingHistory
            timestamp = self.safe_integer_product(item, 'timestamp', 1000)
            type = 'trade'
            amountStr = self.safe_string(item, 'realizedPnl')
            if Precise.string_lt(amountStr, '0'):
                direction = 'out'
                amountStr = Precise.string_neg(amountStr)
            else:
                direction = 'in'
            amount = self.parse_number(amountStr)
            fee = {'currency': 'USDT', 'cost': self.safe_number(item, 'fee')}
            status = 'ok'
        else:
            # response from CapitalHistory
            timestamp = self.safe_integer_product(item, 'timestampSec', 1000)
            amount = self.safe_number(item, 'quantity')
            direction = 'in' if (transactionType == 'deposit' or transactionType == 'transfer-in') else 'out'
            type = self.parse_transaction_type(transactionType)
            status = self.parse_transaction_status(self.safe_string(item, 'status'))
            if transactionType == 'transfer-in':
                referenceAccount = self.safe_string(item, 'srcAccountId')
            elif transactionType == 'transfer-out':
                referenceAccount = self.safe_string(item, 'receivingAccountId')
            referenceId = self.safe_string(item, 'transactionHash')
        return self.safe_ledger_entry({
            'id': self.safe_string(item, 'id'),
            'currency': self.currency('USDT'),
            'account': self.number_to_string(self.accountId),
            'referenceAccount': referenceAccount,
            'referenceId': referenceId,
            'status': status,
            'amount': amount,
            'before': None,
            'after': None,
            'fee': fee,
            'direction': direction,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'type': type,
            'info': item,
        }, currency)

    def fetch_ledger(self, code: Str = None, since: Int = None, limit: Int = None, params={}) -> List[LedgerEntry]:
        """
        fetch the history of changes, actions done by the user or operations that altered the balance of the user

        https://api-doc.hibachi.xyz/#35125e3f-d154-4bfd-8276-a48bb1c62020

        :param str [code]: unified currency code, default is None
        :param int [since]: timestamp in ms of the earliest ledger entry, default is None
        :param int [limit]: max number of ledger entries to return, default is None
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ledger structure <https://docs.ccxt.com/#/?id=ledger>`
        """
        self.load_markets()
        currency = self.currency('USDT')
        request = {'accountId': self.get_account_id()}
        rawPromises = [
            self.privateGetCapitalHistory(self.extend(request, params)),
            self.privateGetTradeAccountTradingHistory(self.extend(request, params)),
        ]
        promises = rawPromises
        responseCapitalHistory = promises[0]
        #
        # {
        #     "transactions": [
        #         {
        #             "assetId": 1,
        #             "blockNumber": 358396669,
        #             "chain": "Arbitrum",
        #             "etaTsSec": null,
        #             "id": 358396669,
        #             "quantity": "0.999500",
        #             "status": "pending",
        #             "timestampSec": 1752692872,
        #             "token": "USDT",
        #             "transactionHash": "0x408e48881e0ba77d8638e3fe57bc06bdec513ddaa8b672e0aefa7e22e2f18b5e",
        #             "transactionType": "deposit"
        #         },
        #         {
        #             "assetId": 1,
        #             "etaTsSec": null,
        #             "id": 13116,
        #             "instantWithdrawalChain": null,
        #             "instantWithdrawalToken": null,
        #             "isInstantWithdrawal": False,
        #             "quantity": "0.040000",
        #             "status": "completed",
        #             "timestampSec": 1752542708,
        #             "transactionHash": "0xe89cf90b2408d1a273dc9427654145def102d9449e5e2cfc10690ccffc3d7e28",
        #             "transactionType": "withdrawal",
        #             "withdrawalAddress": "0x23625d5fc6a6e32638d908eb4c3a3415e5121f76"
        #         },
        #         {
        #             "assetId": 1,
        #             "id": 167,
        #             "quantity": "10.000000",
        #             "srcAccountId": 175,
        #             "srcAddress": "0xc2f77ce029438a3fdfe68ddee25991a9fb985a86",
        #             "status": "completed",
        #             "timestampSec": 1732224729,
        #             "transactionType": "transfer-in"
        #         },
        #         {
        #             "assetId": 1,
        #             "id": 170,
        #             "quantity": "10.000000",
        #             "receivingAccountId": 175,
        #             "receivingAddress": "0xc2f77ce029438a3fdfe68ddee25991a9fb985a86",
        #             "status": "completed",
        #             "timestampSec": 1732225631,
        #             "transactionType": "transfer-out"
        #         },
        #     ]
        # }
        #
        rowsCapitalHistory = self.safe_list(responseCapitalHistory, 'transactions')
        responseTradingHistory = promises[1]
        #
        # {
        #     "tradingHistory": [
        #         {
        #             "eventType": "MARKET",
        #             "fee": "0.000008",
        #             "priceOrFundingRate": "119687.82481",
        #             "quantity": "0.0000003727",
        #             "realizedPnl": "0.004634",
        #             "side": "Sell",
        #             "symbol": "BTC/USDT-P",
        #             "timestamp": 1752522571
        #         },
        #         {
        #             "eventType": "FundingEvent",
        #             "fee": "0",
        #             "priceOrFundingRate": "0.000203",
        #             "quantity": "0.0000003727",
        #             "realizedPnl": "-0.000009067899008751979",
        #             "side": "Long",
        #             "symbol": "BTC/USDT-P",
        #             "timestamp": 1752508800
        #         },
        #     ]
        # }
        #
        rowsTradingHistory = self.safe_list(responseTradingHistory, 'tradingHistory')
        rows = self.array_concat(rowsCapitalHistory, rowsTradingHistory)
        return self.parse_ledger(rows, currency, since, limit, params)

    def fetch_deposit_address(self, code: str, params={}) -> DepositAddress:
        """
        fetch deposit address for given currency and chain. currently, we have a single EVM address across multiple EVM chains. Note: This method is currently only supported for trustless accounts
        :param str code: unified currency code
        :param dict [params]: extra parameters for API
        :param str [params.publicKey]: your public key, you can get it from UI after creating API key
        :returns dict: an `address structure <https://docs.ccxt.com/#/?id=address-structure>`
        """
        request = {
            'publicKey': self.safe_string(params, 'publicKey'),
            'accountId': self.get_account_id(),
        }
        response = self.privateGetCapitalDepositInfo(self.extend(request, params))
        # {
        #     "depositAddressEvm": "0x0b95d90b9345dadf1460bd38b9f4bb0d2f4ed788"
        # }
        return {
            'info': response,
            'currency': 'USDT',
            'network': 'ARBITRUM',
            'address': self.safe_string(response, 'depositAddressEvm'),
            'tag': None,
        }

    def parse_transaction(self, transaction: dict, currency: Currency = None) -> Transaction:
        timestamp = self.safe_integer_product(transaction, 'timestampSec', 1000)
        address = self.safe_string(transaction, 'withdrawalAddress')
        transactionType = self.safe_string(transaction, 'transactionType')
        if transactionType != 'deposit' and transactionType != 'withdrawal':
            transactionType = self.parse_transaction_type(transactionType)
        return {
            'info': transaction,
            'id': self.safe_string(transaction, 'id'),
            'txid': self.safe_string(transaction, 'transactionHash'),
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'network': 'ARBITRUM',  # Currently the exchange only exists on Arbitrum,
            'address': address,
            'addressTo': address,
            'addressFrom': None,
            'tag': None,
            'tagTo': None,
            'tagFrom': None,
            'type': transactionType,
            'amount': self.safe_number(transaction, 'quantity'),
            'currency': 'USDT',
            'status': self.parse_transaction_status(self.safe_string(transaction, 'status')),
            'updated': None,
            'internal': None,
            'comment': None,
            'fee': None,
        }

    def fetch_deposits(self, code: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Transaction]:
        """
        fetch deposits made to account

        https://api-doc.hibachi.xyz/#35125e3f-d154-4bfd-8276-a48bb1c62020

        :param str [code]: unified currency code
        :param int [since]: filter by earliest timestamp(ms)
        :param int [limit]: maximum number of deposits to be returned
        :param dict [params]: extra parameters to be passed to API
        :returns dict[]: a list of `transaction structures <https://docs.ccxt.com/#/?id=transaction-structure>`
        """
        currency = self.safe_currency(code)
        request = {
            'accountId': self.get_account_id(),
        }
        response = self.privateGetCapitalHistory(self.extend(request, params))
        # {
        #     "transactions": [
        #         {
        #             "assetId": 1,
        #             "blockNumber": 0,
        #             "chain": null,
        #             "etaTsSec": 1752758789,
        #             "id": 42688,
        #             "quantity": "6.130000",
        #             "status": "completed",
        #             "timestampSec": 1752758788,
        #             "token": null,
        #             "transactionHash": "0x8dcd7bd1155b5624fb5e38a1365888f712ec633a57434340e05080c70b0e3bba",
        #             "transactionType": "deposit"
        #         },
        #         {
        #             "assetId": 1,
        #             "etaTsSec": null,
        #             "id": 12993,
        #             "instantWithdrawalChain": null,
        #             "instantWithdrawalToken": null,
        #             "isInstantWithdrawal": False,
        #             "quantity": "0.111930",
        #             "status": "completed",
        #             "timestampSec": 1752387891,
        #             "transactionHash": "0x32ab5fe5b90f6d753bab83523ebc8465eb9daef54580e13cb9ff031d400c5620",
        #             "transactionType": "withdrawal",
        #             "withdrawalAddress": "0x43f15ef2ef2ab5e61e987ee3d652a5872aea8a6c"
        #         },
        #     ]
        # }
        transactions = self.safe_list(response, 'transactions')
        deposits = []
        for i in range(0, len(transactions)):
            transaction = transactions[i]
            if self.safe_string(transaction, 'transactionType') == 'deposit':
                deposits.append(transaction)
        return self.parse_transactions(deposits, currency, since, limit, params)

    def fetch_withdrawals(self, code: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Transaction]:
        """
        fetch withdrawals made from account

        https://api-doc.hibachi.xyz/#35125e3f-d154-4bfd-8276-a48bb1c62020

        :param str [code]: unified currency code
        :param int [since]: filter by earliest timestamp(ms)
        :param int [limit]: maximum number of deposits to be returned
        :param dict [params]: extra parameters to be passed to API
        :returns dict[]: a list of `transaction structures <https://docs.ccxt.com/#/?id=transaction-structure>`
        """
        currency = self.safe_currency(code)
        request = {
            'accountId': self.get_account_id(),
        }
        response = self.privateGetCapitalHistory(self.extend(request, params))
        # {
        #     "transactions": [
        #         {
        #             "assetId": 1,
        #             "blockNumber": 0,
        #             "chain": null,
        #             "etaTsSec": 1752758789,
        #             "id": 42688,
        #             "quantity": "6.130000",
        #             "status": "completed",
        #             "timestampSec": 1752758788,
        #             "token": null,
        #             "transactionHash": "0x8dcd7bd1155b5624fb5e38a1365888f712ec633a57434340e05080c70b0e3bba",
        #             "transactionType": "deposit"
        #         },
        #         {
        #             "assetId": 1,
        #             "etaTsSec": null,
        #             "id": 12993,
        #             "instantWithdrawalChain": null,
        #             "instantWithdrawalToken": null,
        #             "isInstantWithdrawal": False,
        #             "quantity": "0.111930",
        #             "status": "completed",
        #             "timestampSec": 1752387891,
        #             "transactionHash": "0x32ab5fe5b90f6d753bab83523ebc8465eb9daef54580e13cb9ff031d400c5620",
        #             "transactionType": "withdrawal",
        #             "withdrawalAddress": "0x43f15ef2ef2ab5e61e987ee3d652a5872aea8a6c"
        #         },
        #     ]
        # }
        transactions = self.safe_list(response, 'transactions')
        withdrawals = []
        for i in range(0, len(transactions)):
            transaction = transactions[i]
            if self.safe_string(transaction, 'transactionType') == 'withdrawal':
                withdrawals.append(transaction)
        return self.parse_transactions(withdrawals, currency, since, limit, params)

    def fetch_time(self, params={}) -> Int:
        """
        fetches the current integer timestamp in milliseconds from the exchange server

        http://api-doc.hibachi.xyz/#b5c6a3bc-243d-4d35-b6d4-a74c92495434

        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int: the current integer timestamp in milliseconds from the exchange server
        """
        response = self.publicGetExchangeUtcTimestamp(params)
        #
        #     {"timestampMs":1754077574040}
        #
        return self.safe_integer(response, 'timestampMs')

    def fetch_open_interest(self, symbol: str, params={}):
        """
        retrieves the open interest of a contract trading pair

        https://api-doc.hibachi.xyz/#bc34e8ae-e094-4802-8d56-3efe3a7bad49

        :param str symbol: unified CCXT market symbol
        :param dict [params]: exchange specific parameters
        :returns dict} an open interest structure{@link https://docs.ccxt.com/#/?id=open-interest-structure:
        """
        self.load_markets()
        market = self.market(symbol)
        request: dict = {
            'symbol': market['id'],
        }
        response = self.publicGetMarketDataOpenInterest(self.extend(request, params))
        #
        #   {"totalQuantity" : "2.3299770166"}
        #
        timestamp = self.milliseconds()
        return self.safe_open_interest({
            'symbol': symbol,
            'openInterestAmount': self.safe_string(response, 'totalQuantity'),
            'openInterestValue': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'info': response,
        }, market)

    def fetch_funding_rate(self, symbol: str, params={}) -> FundingRate:
        """
        fetch the current funding rate

        https://api-doc.hibachi.xyz/#bca696ca-b9b2-4072-8864-5d6b8c09807e

        :param str symbol: unified market symbol
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `funding rate structure <https://docs.ccxt.com/#/?id=funding-rate-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request: dict = {
            'symbol': market['id'],
        }
        response = self.publicGetMarketDataPrices(self.extend(request, params))
        #
        # {
        #     "askPrice": "3514.650296",
        #     "bidPrice": "3513.596112",
        #     "fundingRateEstimation": {
        #         "estimatedFundingRate": "0.000001",
        #         "nextFundingTimestamp": 1712707200
        #     },
        #     "markPrice": "3514.288858",
        #     "spotPrice": "3514.715000",
        #     "symbol": "ETH/USDT-P",
        #     "tradePrice": "2372.746570"
        # }
        #
        funding = self.safe_dict(response, 'fundingRateEstimation', {})
        timestamp = self.milliseconds()
        nextFundingTimestamp = self.safe_integer_product(funding, 'nextFundingTimestamp', 1000)
        return {
            'info': funding,
            'symbol': market['symbol'],
            'markPrice': None,
            'indexPrice': None,
            'interestRate': self.parse_number('0'),
            'estimatedSettlePrice': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'fundingRate': self.safe_number(funding, 'estimatedFundingRate'),
            'fundingTimestamp': nextFundingTimestamp,
            'fundingDatetime': self.iso8601(nextFundingTimestamp),
            'nextFundingRate': None,
            'nextFundingTimestamp': None,
            'nextFundingDatetime': None,
            'previousFundingRate': None,
            'previousFundingTimestamp': None,
            'previousFundingDatetime': None,
            'interval': '8h',
        }

    def fetch_funding_rate_history(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}):
        """
        fetches historical funding rate prices

        https://api-doc.hibachi.xyz/#4abb30c4-e5c7-4b0f-9ade-790111dbfa47

        :param str symbol: unified symbol of the market to fetch the funding rate history for
        :param int [since]: timestamp in ms of the earliest funding rate to fetch
        :param int [limit]: the maximum amount of `funding rate structures <https://docs.ccxt.com/#/?id=funding-rate-history-structure>` to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `funding rate structures <https://docs.ccxt.com/#/?id=funding-rate-history-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        request: dict = {
            'symbol': market['id'],
        }
        response = self.publicGetMarketDataFundingRates(self.extend(request, params))
        #
        # {
        #     "data": [
        #         {
        #             "contractId": 2,
        #             "fundingTimestamp": 1753488000,
        #             "fundingRate": "0.000137",
        #             "indexPrice": "117623.65010"
        #         }
        #     ]
        # }
        #
        data = self.safe_list(response, 'data')
        rates = []
        for i in range(0, len(data)):
            entry = data[i]
            timestamp = self.safe_integer_product(entry, 'fundingTimestamp', 1000)
            rates.append({
                'info': entry,
                'symbol': symbol,
                'fundingRate': self.safe_number(entry, 'fundingRate'),
                'timestamp': timestamp,
                'datetime': self.iso8601(timestamp),
            })
        sorted = self.sort_by(rates, 'timestamp')
        return self.filter_by_symbol_since_limit(sorted, symbol, since, limit)



from ccxt.base.types import Entry


class ImplicitAPI:
    v1_archive_post = v1ArchivePost = Entry('', ['v1', 'archive'], 'POST', {'cost': 1})
    v1_gateway_get_query = v1GatewayGetQuery = Entry('query', ['v1', 'gateway'], 'GET', {'cost': 1})
    v1_gateway_get_symbols = v1GatewayGetSymbols = Entry('symbols', ['v1', 'gateway'], 'GET', {'cost': 1})
    v1_gateway_get_time = v1GatewayGetTime = Entry('time', ['v1', 'gateway'], 'GET', {'cost': 1})
    v1_gateway_post_query = v1GatewayPostQuery = Entry('query', ['v1', 'gateway'], 'POST', {'cost': 1})
    v1_gateway_post_execute = v1GatewayPostExecute = Entry('execute', ['v1', 'gateway'], 'POST', {'cost': 1})
    v1_trigger_post_execute = v1TriggerPostExecute = Entry('execute', ['v1', 'trigger'], 'POST', {'cost': 1})
    v1_trigger_post_query = v1TriggerPostQuery = Entry('query', ['v1', 'trigger'], 'POST', {'cost': 1})
    v2_archive_get_tickers = v2ArchiveGetTickers = Entry('tickers', ['v2', 'archive'], 'GET', {'cost': 1})
    v2_archive_get_contracts = v2ArchiveGetContracts = Entry('contracts', ['v2', 'archive'], 'GET', {'cost': 1})
    v2_archive_get_trades = v2ArchiveGetTrades = Entry('trades', ['v2', 'archive'], 'GET', {'cost': 1})
    v2_archive_get_vrtx = v2ArchiveGetVrtx = Entry('vrtx', ['v2', 'archive'], 'GET', {'cost': 1})
    v2_gateway_get_assets = v2GatewayGetAssets = Entry('assets', ['v2', 'gateway'], 'GET', {'cost': 0.6667})
    v2_gateway_get_pairs = v2GatewayGetPairs = Entry('pairs', ['v2', 'gateway'], 'GET', {'cost': 1})
    v2_gateway_get_orderbook = v2GatewayGetOrderbook = Entry('orderbook', ['v2', 'gateway'], 'GET', {'cost': 1})



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_otc_v1_market_pair = publicGetOtcV1MarketPair = Entry('otc/v1/market/pair', 'public', 'GET', {'cost': 1})
    public_get_otc_v1_market_order_pair_price = publicGetOtcV1MarketOrderPairPrice = Entry('otc/v1/market/order/pair/price', 'public', 'GET', {'cost': 1})



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_v3_symbols = publicGetV3Symbols = Entry('v3/symbols', 'public', 'GET', {'cost': 1})
    public_get_v3_history_funding = publicGetV3HistoryFunding = Entry('v3/history-funding', 'public', 'GET', {'cost': 1})
    public_get_v3_ticker = publicGetV3Ticker = Entry('v3/ticker', 'public', 'GET', {'cost': 1})
    public_get_v3_klines = publicGetV3Klines = Entry('v3/klines', 'public', 'GET', {'cost': 1})
    public_get_v3_trades = publicGetV3Trades = Entry('v3/trades', 'public', 'GET', {'cost': 1})
    public_get_v3_depth = publicGetV3Depth = Entry('v3/depth', 'public', 'GET', {'cost': 1})
    public_get_v3_time = publicGetV3Time = Entry('v3/time', 'public', 'GET', {'cost': 1})
    public_get_v3_data_all_ticker_info = publicGetV3DataAllTickerInfo = Entry('v3/data/all-ticker-info', 'public', 'GET', {'cost': 1})
    private_get_v3_account = privateGetV3Account = Entry('v3/account', 'private', 'GET', {'cost': 1})
    private_get_v3_account_balance = privateGetV3AccountBalance = Entry('v3/account-balance', 'private', 'GET', {'cost': 1})
    private_get_v3_fills = privateGetV3Fills = Entry('v3/fills', 'private', 'GET', {'cost': 1})
    private_get_v3_order_fills = privateGetV3OrderFills = Entry('v3/order-fills', 'private', 'GET', {'cost': 1})
    private_get_v3_order = privateGetV3Order = Entry('v3/order', 'private', 'GET', {'cost': 1})
    private_get_v3_history_orders = privateGetV3HistoryOrders = Entry('v3/history-orders', 'private', 'GET', {'cost': 1})
    private_get_v3_order_by_client_order_id = privateGetV3OrderByClientOrderId = Entry('v3/order-by-client-order-id', 'private', 'GET', {'cost': 1})
    private_get_v3_funding = privateGetV3Funding = Entry('v3/funding', 'private', 'GET', {'cost': 1})
    private_get_v3_historical_pnl = privateGetV3HistoricalPnl = Entry('v3/historical-pnl', 'private', 'GET', {'cost': 1})
    private_get_v3_open_orders = privateGetV3OpenOrders = Entry('v3/open-orders', 'private', 'GET', {'cost': 1})
    private_get_v3_transfers = privateGetV3Transfers = Entry('v3/transfers', 'private', 'GET', {'cost': 1})
    private_get_v3_transfer = privateGetV3Transfer = Entry('v3/transfer', 'private', 'GET', {'cost': 1})
    private_post_v3_delete_open_orders = privatePostV3DeleteOpenOrders = Entry('v3/delete-open-orders', 'private', 'POST', {'cost': 1})
    private_post_v3_delete_client_order_id = privatePostV3DeleteClientOrderId = Entry('v3/delete-client-order-id', 'private', 'POST', {'cost': 1})
    private_post_v3_delete_order = privatePostV3DeleteOrder = Entry('v3/delete-order', 'private', 'POST', {'cost': 1})
    private_post_v3_order = privatePostV3Order = Entry('v3/order', 'private', 'POST', {'cost': 1})
    private_post_v3_set_initial_margin_rate = privatePostV3SetInitialMarginRate = Entry('v3/set-initial-margin-rate', 'private', 'POST', {'cost': 1})
    private_post_v3_transfer_out = privatePostV3TransferOut = Entry('v3/transfer-out', 'private', 'POST', {'cost': 1})
    private_post_v3_contract_transfer_out = privatePostV3ContractTransferOut = Entry('v3/contract-transfer-out', 'private', 'POST', {'cost': 1})



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCache, ArrayCacheBySymbolById, ArrayCacheBySymbolBySide, ArrayCacheByTimestamp
import hashlib
from ccxt.base.types import Any, Balances, Bool, Int, Market, Order, OrderBook, Position, Str, Strings, Ticker, Tickers, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import AuthenticationError
from ccxt.base.errors import ArgumentsRequired
from ccxt.base.errors import BadRequest
from ccxt.base.errors import NotSupported
from ccxt.base.errors import RateLimitExceeded
from ccxt.base.errors import ChecksumError
from ccxt.base.errors import UnsubscribeError
from ccxt.base.precise import Precise


class coincatch(ccxt.async_support.coincatch):

    def describe(self) -> Any:
        return self.deep_extend(super(coincatch, self).describe(), {
            'has': {
                'ws': True,
                'watchTrades': True,
                'watchTradesForSymbols': True,
                'watchOrderBook': True,
                'watchOrderBookForSymbols': True,
                'watchOHLCV': True,
                'watchOHLCVForSymbols': False,  # todo
                'watchOrders': True,
                'watchMyTrades': False,
                'watchTicker': True,
                'watchTickers': True,
                'watchBalance': True,
                'watchPositions': True,
            },
            'urls': {
                'api': {
                    'ws': {
                        'public': 'wss://ws.coincatch.com/public/v1/stream',
                        'private': 'wss://ws.coincatch.com/private/v1/stream',
                    },
                },
            },
            'options': {
                'tradesLimit': 1000,
                'OHLCVLimit': 200,
                'timeframesForWs': {
                    '1m': '1m',
                    '5m': '5m',
                    '15m': '15m',
                    '30m': '30m',
                    '1h': '1H',
                    '4h': '4H',
                    '12h': '12H',
                    '1d': '1D',
                    '1w': '1W',
                },
                'watchOrderBook': {
                    'checksum': True,
                },
            },
            'streaming': {
                'ping': self.ping,
            },
            'exceptions': {
                'ws': {
                    'exact': {
                        '30001': BadRequest,  # Channel does not exist
                        '30002': AuthenticationError,  # illegal request
                        '30003': BadRequest,  # invalid op
                        '30004': AuthenticationError,  # User needs to log in
                        '30005': AuthenticationError,  # login failed
                        '30006': RateLimitExceeded,  # request too many
                        '30007': RateLimitExceeded,  # request over limit,connection close
                        '30011': AuthenticationError,  # invalid ACCESS_KEY
                        '30012': AuthenticationError,  # invalid ACCESS_PASSPHRASE
                        '30013': AuthenticationError,  # invalid ACCESS_TIMESTAMP
                        '30014': BadRequest,  # Request timestamp expired
                        '30015': AuthenticationError,  # {event: 'error', code: 30015, msg: 'Invalid sign'}
                    },
                    'broad': {},
                },
            },
        })

    def get_market_from_arg(self, entry):
        instId = self.safe_string(entry, 'instId')
        instType = self.safe_string(entry, 'instType')
        baseAndQuote = self.parseSpotMarketId(instId)
        baseId = baseAndQuote['baseId']
        quoteId = baseAndQuote['quoteId']
        suffix = '_SPBL'  # spot suffix
        if instType == 'mc':
            if quoteId == 'USD':
                suffix = '_DMCBL'
            else:
                suffix = '_UMCBL'
        marketId = self.safe_currency_code(baseId) + self.safe_currency_code(quoteId) + suffix
        return self.safeMarketCustom(marketId)

    async def authenticate(self, params={}):
        self.check_required_credentials()
        url = self.urls['api']['ws']['private']
        client = self.client(url)
        messageHash = 'authenticated'
        future = client.future(messageHash)
        authenticated = self.safe_value(client.subscriptions, messageHash)
        if authenticated is None:
            timestamp = str(self.seconds())
            auth = timestamp + 'GET' + '/user/verify'
            signature = self.hmac(self.encode(auth), self.encode(self.secret), hashlib.sha256, 'base64')
            operation = 'login'
            request: dict = {
                'op': operation,
                'args': [
                    {
                        'apiKey': self.apiKey,
                        'passphrase': self.password,
                        'timestamp': timestamp,
                        'sign': signature,
                    },
                ],
            }
            message = self.extend(request, params)
            self.watch(url, messageHash, message, messageHash)
        return await future

    async def watch_public(self, messageHash, subscribeHash, args, params={}):
        url = self.urls['api']['ws']['public']
        request: dict = {
            'op': 'subscribe',
            'args': [args],
        }
        message = self.extend(request, params)
        return await self.watch(url, messageHash, message, subscribeHash)

    async def un_watch_public(self, messageHash, args, params={}):
        url = self.urls['api']['ws']['public']
        request: dict = {
            'op': 'unsubscribe',
            'args': [args],
        }
        message = self.extend(request, params)
        return await self.watch(url, messageHash, message, messageHash)

    async def watch_private(self, messageHash, subscribeHash, args, params={}):
        await self.authenticate()
        url = self.urls['api']['ws']['private']
        request: dict = {
            'op': 'subscribe',
            'args': [args],
        }
        message = self.extend(request, params)
        return await self.watch(url, messageHash, message, subscribeHash)

    async def watch_private_multiple(self, messageHashes, subscribeHashes, args, params={}):
        await self.authenticate()
        url = self.urls['api']['ws']['private']
        request: dict = {
            'op': 'subscribe',
            'args': args,
        }
        message = self.extend(request, params)
        return await self.watch_multiple(url, messageHashes, message, subscribeHashes)

    def handle_authenticate(self, client: Client, message):
        #
        #  {event: "login", code: 0}
        #
        messageHash = 'authenticated'
        future = self.safe_value(client.futures, messageHash)
        future.resolve(True)

    async def watch_public_multiple(self, messageHashes, subscribeHashes, argsArray, params={}):
        url = self.urls['api']['ws']['public']
        request: dict = {
            'op': 'subscribe',
            'args': argsArray,
        }
        message = self.extend(request, params)
        return await self.watch_multiple(url, messageHashes, message, subscribeHashes)

    async def un_watch_channel(self, symbol: str, channel: str, messageHashTopic: str, params={}) -> Any:
        await self.load_markets()
        market = self.market(symbol)
        instType, instId = self.get_public_inst_type_and_id(market)
        messageHash = 'unsubscribe:' + messageHashTopic + ':' + symbol
        args: dict = {
            'instType': instType,
            'channel': channel,
            'instId': instId,
        }
        return await self.un_watch_public(messageHash, args, params)

    def get_public_inst_type_and_id(self, market: Market):
        instId = market['baseId'] + market['quoteId']
        instType = None
        if market['spot']:
            instType = 'SP'
        elif market['swap']:
            instType = 'MC'
        else:
            raise NotSupported(self.id + ' supports only spot and swap markets')
        return [instType, instId]

    def handle_dmcbl_market_by_message_hashes(self, market: Market, hash: str, client: Client, timeframe: Str = None):
        marketId = market['id']
        messageHashes = self.find_message_hashes(client, hash)
        # the exchange counts DMCBL markets same market with different quote currencies
        # for example symbols ETHUSD:ETH and ETH/USD:BTC both have the same marketId ETHUSD_DMCBL
        # we need to check all markets with the same marketId to find the correct market that is in messageHashes
        marketsWithCurrentId = self.safe_list(self.markets_by_id, marketId, [])
        suffix = ''
        if timeframe is not None:
            suffix = ':' + timeframe
        for i in range(0, len(marketsWithCurrentId)):
            market = marketsWithCurrentId[i]
            symbol = market['symbol']
            messageHash = hash + symbol + suffix
            if self.in_array(messageHash, messageHashes):
                return market
        return market

    async def watch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market

        https://coincatch.github.io/github.io/en/spot/#tickers-channel

        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.instType]: the type of the instrument to fetch the ticker for, 'SP' for spot markets, 'MC' for futures markets(default is 'SP')
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        instType, instId = self.get_public_inst_type_and_id(market)
        channel = 'ticker'
        messageHash = channel + ':' + symbol
        args: dict = {
            'instType': instType,
            'channel': channel,
            'instId': instId,
        }
        return await self.watch_public(messageHash, messageHash, args, params)

    async def un_watch_ticker(self, symbol: str, params={}) -> Any:
        """
        unsubscribe from the ticker channel

        https://coincatch.github.io/github.io/en/mix/#tickers-channel

        :param str symbol: unified symbol of the market to unwatch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns any: status of the unwatch request
        """
        await self.load_markets()
        return await self.un_watch_channel(symbol, 'ticker', 'ticker', params)

    async def watch_tickers(self, symbols: Strings = None, params={}) -> Tickers:
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for all markets of a specific list

        https://coincatch.github.io/github.io/en/mix/#tickers-channel

        :param str[] symbols: unified symbol of the market to watch the tickers for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        if symbols is None:
            symbols = self.symbols
        topics = []
        messageHashes = []
        for i in range(0, len(symbols)):
            symbol = symbols[i]
            market = self.market(symbol)
            instType, instId = self.get_public_inst_type_and_id(market)
            args: dict = {
                'instType': instType,
                'channel': 'ticker',
                'instId': instId,
            }
            topics.append(args)
            messageHashes.append('ticker:' + symbol)
        tickers = await self.watch_public_multiple(messageHashes, messageHashes, topics, params)
        if self.newUpdates:
            result: dict = {}
            result[tickers['symbol']] = tickers
            return result
        return self.filter_by_array(self.tickers, 'symbol', symbols)

    def handle_ticker(self, client: Client, message):
        #
        #     action: 'snapshot',
        #     arg: {instType: 'sp', channel: 'ticker', instId: 'ETHUSDT'},
        #     data: [
        #         {
        #             instId: 'ETHUSDT',
        #             last: '2421.06',
        #             open24h: '2416.93',
        #             high24h: '2441.47',
        #             low24h: '2352.99',
        #             bestBid: '2421.03',
        #             bestAsk: '2421.06',
        #             baseVolume: '9445.2043',
        #             quoteVolume: '22807159.1148',
        #             ts: 1728131730687,
        #             labeId: 0,
        #             openUtc: '2414.50',
        #             chgUTC: '0.00272',
        #             bidSz: '3.866',
        #             askSz: '0.124'
        #         }
        #     ],
        #     ts: 1728131730688
        #
        arg = self.safe_dict(message, 'arg', {})
        market = self.get_market_from_arg(arg)
        marketId = market['id']
        hash = 'ticker:'
        if marketId.find('_DMCBL') >= 0:
            market = self.handle_dmcbl_market_by_message_hashes(market, hash, client)
        data = self.safe_list(message, 'data', [])
        ticker = self.parse_ws_ticker(self.safe_dict(data, 0, {}), market)
        symbol = market['symbol']
        self.tickers[symbol] = ticker
        messageHash = hash + symbol
        client.resolve(self.tickers[symbol], messageHash)

    def parse_ws_ticker(self, ticker, market=None):
        #
        # spot
        #     {
        #         instId: 'ETHUSDT',
        #         last: '2421.06',
        #         open24h: '2416.93',
        #         high24h: '2441.47',
        #         low24h: '2352.99',
        #         bestBid: '2421.03',
        #         bestAsk: '2421.06',
        #         baseVolume: '9445.2043',
        #         quoteVolume: '22807159.1148',
        #         ts: 1728131730687,
        #         labeId: 0,
        #         openUtc: '2414.50',
        #         chgUTC: '0.00272',
        #         bidSz: '3.866',
        #         askSz: '0.124'
        #     }
        #
        # swap
        #     {
        #         instId: 'ETHUSDT',
        #         last: '2434.47',
        #         bestAsk: '2434.48',
        #         bestBid: '2434.47',
        #         high24h: '2471.68',
        #         low24h: '2400.01',
        #         priceChangePercent: '0.00674',
        #         capitalRate: '0.000082',
        #         nextSettleTime: 1728489600000,
        #         systemTime: 1728471993602,
        #         markPrice: '2434.46',
        #         indexPrice: '2435.44',
        #         holding: '171450.25',
        #         baseVolume: '1699298.91',
        #         quoteVolume: '4144522832.32',
        #         openUtc: '2439.67',
        #         chgUTC: '-0.00213',
        #         symbolType: 1,
        #         symbolId: 'ETHUSDT_UMCBL',
        #         deliveryPrice: '0',
        #         bidSz: '26.12',
        #         askSz: '49.6'
        #     }
        #
        last = self.safe_string(ticker, 'last')
        timestamp = self.safe_integer_2(ticker, 'ts', 'systemTime')
        return self.safe_ticker({
            'symbol': market['symbol'],
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'high': self.safe_string(ticker, 'high24h'),
            'low': self.safe_string(ticker, 'low24h'),
            'bid': self.safe_string(ticker, 'bestBid'),
            'bidVolume': self.safe_string(ticker, 'bidSz'),
            'ask': self.safe_string(ticker, 'bestAsk'),
            'askVolume': self.safe_string(ticker, 'askSz'),
            'vwap': None,
            'open': self.safe_string_2(ticker, 'open24h', 'openUtc'),
            'close': last,
            'last': last,
            'previousClose': None,
            'change': None,
            'percentage': Precise.string_mul(self.safe_string(ticker, 'chgUTC'), '100'),
            'average': None,
            'baseVolume': self.safe_number(ticker, 'baseVolume'),
            'quoteVolume': self.safe_number(ticker, 'quoteVolume'),
            'indexPrice': self.safe_string(ticker, 'indexPrice'),
            'markPrice': self.safe_string(ticker, 'markPrice'),
            'info': ticker,
        }, market)

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market

        https://coincatch.github.io/github.io/en/spot/#candlesticks-channel

        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch(not including)
        :param int [limit]: the maximum amount of candles to fetch(not including)
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param bool [params.instType]: the type of the instrument to fetch the OHLCV data for, 'SP' for spot markets, 'MC' for futures markets(default is 'SP')
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        market = self.market(symbol)
        timeframes = self.options['timeframesForWs']
        channel = 'candle' + self.safe_string(timeframes, timeframe)
        instType, instId = self.get_public_inst_type_and_id(market)
        args: dict = {
            'instType': instType,
            'channel': channel,
            'instId': instId,
        }
        messageHash = 'ohlcv:' + symbol + ':' + timeframe
        ohlcv = await self.watch_public(messageHash, messageHash, args, params)
        if self.newUpdates:
            limit = ohlcv.getLimit(symbol, limit)
        return self.filter_by_since_limit(ohlcv, since, limit, 0, True)

    async def un_watch_ohlcv(self, symbol: str, timeframe='1m', params={}) -> Any:
        """
        unsubscribe from the ohlcv channel

        https://www.bitget.com/api-doc/spot/websocket/public/Candlesticks-Channel

        :param str symbol: unified symbol of the market to unwatch the ohlcv for
 @param timeframe
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        timeframes = self.options['timeframesForWs']
        interval = self.safe_string(timeframes, timeframe)
        channel = 'candle' + interval
        return await self.un_watch_channel(symbol, channel, 'ohlcv:' + interval, params)

    def handle_ohlcv(self, client: Client, message):
        #
        #     {
        #         action: 'update',
        #         arg: {instType: 'sp', channel: 'candle1D', instId: 'ETHUSDT'},
        #         data: [
        #             [
        #                 '1728316800000',
        #                 '2474.5',
        #                 '2478.21',
        #                 '2459.8',
        #                 '2463.51',
        #                 '86.0551'
        #             ]
        #         ],
        #         ts: 1728317607657
        #     }
        #
        arg = self.safe_dict(message, 'arg', {})
        market = self.get_market_from_arg(arg)
        marketId = market['id']
        hash = 'ohlcv:'
        data = self.safe_list(message, 'data', [])
        channel = self.safe_string(arg, 'channel')
        klineType = channel[6:]
        timeframe = self.find_timeframe(klineType)
        if marketId.find('_DMCBL') >= 0:
            market = self.handle_dmcbl_market_by_message_hashes(market, hash, client, timeframe)
        symbol = market['symbol']
        if not (symbol in self.ohlcvs):
            self.ohlcvs[symbol] = {}
        if not (timeframe in self.ohlcvs[symbol]):
            limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
            self.ohlcvs[symbol][timeframe] = ArrayCacheByTimestamp(limit)
        stored = self.ohlcvs[symbol][timeframe]
        for i in range(0, len(data)):
            candle = self.safe_list(data, i, [])
            parsed = self.parse_ws_ohlcv(candle, market)
            stored.append(parsed)
        messageHash = hash + symbol + ':' + timeframe
        client.resolve(stored, messageHash)

    def parse_ws_ohlcv(self, ohlcv, market: Market = None) -> list:
        #
        #     [
        #         '1728316800000',
        #         '2474.5',
        #         '2478.21',
        #         '2459.8',
        #         '2463.51',
        #         '86.0551'
        #     ]
        #
        return [
            self.safe_integer(ohlcv, 0),
            self.safe_number(ohlcv, 1),
            self.safe_number(ohlcv, 2),
            self.safe_number(ohlcv, 3),
            self.safe_number(ohlcv, 4),
            self.safe_number(ohlcv, 5),
        ]

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data

        https://coincatch.github.io/github.io/en/spot/#depth-channel

        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        return await self.watch_order_book_for_symbols([symbol], limit, params)

    async def un_watch_order_book(self, symbol: str, params={}) -> Any:
        """
        unsubscribe from the orderbook channel

        https://coincatch.github.io/github.io/en/spot/#depth-channel

        :param str symbol: unified symbol of the market to fetch the order book for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param int [params.limit]: orderbook limit, default is None
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        channel = 'books'
        limit = self.safe_integer(params, 'limit')
        if (limit == 5) or (limit == 15):
            params = self.omit(params, 'limit')
            channel += str(limit)
        return await self.un_watch_channel(symbol, channel, channel, params)

    async def watch_order_book_for_symbols(self, symbols: List[str], limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data

        https://coincatch.github.io/github.io/en/spot/#depth-channel

 @param symbols
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        symbols = self.market_symbols(symbols)
        channel = 'books'
        topics = []
        messageHashes = []
        for i in range(0, len(symbols)):
            symbol = symbols[i]
            market = self.market(symbol)
            instType, instId = self.get_public_inst_type_and_id(market)
            args: dict = {
                'instType': instType,
                'channel': channel,
                'instId': instId,
            }
            topics.append(args)
            messageHashes.append(channel + ':' + symbol)
        orderbook = await self.watch_public_multiple(messageHashes, messageHashes, topics, params)
        return orderbook.limit()

    def handle_order_book(self, client: Client, message):
        #
        #     {
        #         action: 'update',
        #         arg: {instType: 'sp', channel: 'books', instId: 'ETHUSDT'},
        #         data: [
        #             {
        #                 asks: [[2507.07, 0.4248]],
        #                 bids: [[2507.05, 0.1198]],
        #                 checksum: -1400923312,
        #                 ts: '1728339446908'
        #             }
        #         ],
        #         ts: 1728339446908
        #     }
        #
        arg = self.safe_dict(message, 'arg', {})
        market = self.get_market_from_arg(arg)
        marketId = market['id']
        hash = 'books:'
        if marketId.find('_DMCBL') >= 0:
            market = self.handle_dmcbl_market_by_message_hashes(market, hash, client)
        symbol = market['symbol']
        channel = self.safe_string(arg, 'channel')
        messageHash = hash + symbol
        data = self.safe_list(message, 'data', [])
        rawOrderBook = self.safe_dict(data, 0)
        timestamp = self.safe_integer(rawOrderBook, 'ts')
        incrementalBook = channel
        if incrementalBook:
            if not (symbol in self.orderbooks):
                ob = self.counted_order_book({})
                ob['symbol'] = symbol
                self.orderbooks[symbol] = ob
            storedOrderBook = self.orderbooks[symbol]
            asks = self.safe_list(rawOrderBook, 'asks', [])
            bids = self.safe_list(rawOrderBook, 'bids', [])
            self.handle_deltas(storedOrderBook['asks'], asks)
            self.handle_deltas(storedOrderBook['bids'], bids)
            storedOrderBook['timestamp'] = timestamp
            storedOrderBook['datetime'] = self.iso8601(timestamp)
            checksum = self.safe_bool(self.options, 'checksum', True)
            isSnapshot = self.safe_string(message, 'action') == 'snapshot'
            if not isSnapshot and checksum:
                storedAsks = storedOrderBook['asks']
                storedBids = storedOrderBook['bids']
                asksLength = len(storedAsks)
                bidsLength = len(storedBids)
                payloadArray = []
                for i in range(0, 25):
                    if i < bidsLength:
                        payloadArray.append(storedBids[i][2][0])
                        payloadArray.append(storedBids[i][2][1])
                    if i < asksLength:
                        payloadArray.append(storedAsks[i][2][0])
                        payloadArray.append(storedAsks[i][2][1])
                payload = ':'.join(payloadArray)
                calculatedChecksum = self.crc32(payload, True)
                responseChecksum = self.safe_integer(rawOrderBook, 'checksum')
                if calculatedChecksum != responseChecksum:
                    self.spawn(self.handle_check_sum_error, client, symbol, messageHash)
                    return
        else:
            orderbook = self.order_book({})
            parsedOrderbook = self.parse_order_book(rawOrderBook, symbol, timestamp)
            orderbook.reset(parsedOrderbook)
            self.orderbooks[symbol] = orderbook
        client.resolve(self.orderbooks[symbol], messageHash)

    async def handle_check_sum_error(self, client: Client, symbol: str, messageHash: str):
        await self.un_watch_order_book(symbol)
        error = ChecksumError(self.id + ' ' + self.orderbook_checksum_message(symbol))
        client.reject(error, messageHash)

    def handle_delta(self, bookside, delta):
        bidAsk = self.parse_bid_ask(delta, 0, 1)
        bidAsk.append(delta)
        bookside.storeArray(bidAsk)

    def handle_deltas(self, bookside, deltas):
        for i in range(0, len(deltas)):
            self.handle_delta(bookside, deltas[i])

    async def watch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a particular symbol

        https://coincatch.github.io/github.io/en/spot/#trades-channel

        :param str symbol: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        return await self.watch_trades_for_symbols([symbol], since, limit, params)

    async def watch_trades_for_symbols(self, symbols: List[str], since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        watches information on multiple trades made in a market

        https://coincatch.github.io/github.io/en/spot/#trades-channel

 @param symbols
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of trade structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=trade-structure>`
        """
        symbolsLength = len(symbols)
        if symbolsLength == 0:
            raise ArgumentsRequired(self.id + ' watchTradesForSymbols() requires a non-empty array of symbols')
        await self.load_markets()
        symbols = self.market_symbols(symbols)
        topics = []
        messageHashes = []
        for i in range(0, len(symbols)):
            symbol = symbols[i]
            market = self.market(symbol)
            instType, instId = self.get_public_inst_type_and_id(market)
            args: dict = {
                'instType': instType,
                'channel': 'trade',
                'instId': instId,
            }
            topics.append(args)
            messageHashes.append('trade:' + symbol)
        trades = await self.watch_public_multiple(messageHashes, messageHashes, topics, params)
        if self.newUpdates:
            first = self.safe_dict(trades, 0)
            tradeSymbol = self.safe_string(first, 'symbol')
            limit = trades.getLimit(tradeSymbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp', True)

    async def un_watch_trades(self, symbol: str, params={}) -> Any:
        """
        unsubscribe from the trades channel

        https://coincatch.github.io/github.io/en/spot/#trades-channel

        :param str symbol: unified symbol of the market to unwatch the trades for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns any: status of the unwatch request
        """
        await self.load_markets()
        return await self.un_watch_channel(symbol, 'trade', 'trade', params)

    def handle_trades(self, client: Client, message):
        #
        #     {
        #         action: 'update',
        #         arg: {instType: 'sp', channel: 'trade', instId: 'ETHUSDT'},
        #         data: [['1728341807469', '2421.41', '0.478', 'sell']],
        #         ts: 1728341807482
        #     }
        #
        arg = self.safe_dict(message, 'arg', {})
        market = self.get_market_from_arg(arg)
        marketId = market['id']
        hash = 'trade:'
        if marketId.find('_DMCBL') >= 0:
            market = self.handle_dmcbl_market_by_message_hashes(market, hash, client)
        symbol = market['symbol']
        if not (symbol in self.trades):
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            self.trades[symbol] = ArrayCache(limit)
        stored = self.trades[symbol]
        data = self.safe_list(message, 'data', [])
        if data is not None:
            data = self.sort_by(data, 0)
            for i in range(0, len(data)):
                trade = self.safe_list(data, i)
                parsed = self.parse_ws_trade(trade, market)
                stored.append(parsed)
        messageHash = hash + symbol
        client.resolve(stored, messageHash)

    def parse_ws_trade(self, trade, market=None) -> Trade:
        #
        #     [
        #         '1728341807469',
        #         '2421.41',
        #         '0.478',
        #         'sell'
        #     ]
        #
        timestamp = self.safe_integer(trade, 0)
        return self.safe_trade({
            'id': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': market['symbol'],
            'side': self.safe_string_lower(trade, 3),
            'price': self.safe_string(trade, 1),
            'amount': self.safe_string(trade, 2),
            'cost': None,
            'takerOrMaker': None,
            'type': None,
            'order': None,
            'fee': None,
            'info': trade,
        }, market)

    async def watch_balance(self, params={}) -> Balances:
        """
        watch balance and get the amount of funds available for trading or funds locked in orders

        https://coincatch.github.io/github.io/en/spot/#account-channel
        https://coincatch.github.io/github.io/en/mix/#account-channel

        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.type]: 'spot' or 'swap'(default is 'spot')
        :param str [params.instType]: *swap only* 'umcbl' or 'dmcbl'(default is 'umcbl')
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        type = None
        type, params = self.handle_market_type_and_params('watchBalance', None, params)
        instType = 'spbl'  # must be lower case for spot
        if type == 'swap':
            instType = 'umcbl'
        channel = 'account'
        instType, params = self.handle_option_and_params(params, 'watchBalance', 'instType', instType)
        args: dict = {
            'instType': instType,
            'channel': channel,
            'instId': 'default',
        }
        messageHash = 'balance:' + instType.lower()
        return await self.watch_private(messageHash, messageHash, args, params)

    def handle_balance(self, client: Client, message):
        #
        #  spot
        #     {
        #         action: 'snapshot',
        #         arg: {instType: 'spbl', channel: 'account', instId: 'default'},
        #         data: [
        #             {
        #                 coinId: '3',
        #                 coinName: 'ETH',
        #                 available: '0.0000832',
        #                 frozen: '0',
        #                 lock: '0'
        #             }
        #         ],
        #         ts: 1728464548725
        #     }
        #
        #  #  swap
        #     {
        #         action: 'snapshot',
        #         arg: {instType: 'dmcbl', channel: 'account', instId: 'default'},
        #         data: [
        #             {
        #                 marginCoin: 'ETH',
        #                 locked: '0.00000000',
        #                 available: '0.00001203',
        #                 maxOpenPosAvailable: '0.00001203',
        #                 maxTransferOut: '0.00001203',
        #                 equity: '0.00001203',
        #                 usdtEquity: '0.029092328738',
        #                 coinDisplayName: 'ETH'
        #             }
        #         ],
        #         ts: 1728650777643
        #     }
        #
        data = self.safe_list(message, 'data', [])
        for i in range(0, len(data)):
            rawBalance = data[i]
            currencyId = self.safe_string_2(rawBalance, 'coinName', 'marginCoin')
            code = self.safe_currency_code(currencyId)
            account = self.balance[code] if (code in self.balance) else self.account()
            freeQuery = 'maxTransferOut' if ('maxTransferOut' in rawBalance) else 'available'
            account['free'] = self.safe_string(rawBalance, freeQuery)
            account['total'] = self.safe_string(rawBalance, 'equity')
            account['used'] = self.safe_string(rawBalance, 'frozen')
            self.balance[code] = account
        self.balance = self.safe_balance(self.balance)
        arg = self.safe_dict(message, 'arg')
        instType = self.safe_string_lower(arg, 'instType')
        messageHash = 'balance:' + instType
        client.resolve(self.balance, messageHash)

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        watches information on multiple orders made by the user

        https://coincatch.github.io/github.io/en/spot/#order-channel
        https://coincatch.github.io/github.io/en/mix/#order-channel
        https://coincatch.github.io/github.io/en/mix/#plan-order-channel

        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.type]: 'spot' or 'swap'
        :param str [params.instType]: *swap only* 'umcbl' or 'dmcbl'(default is 'umcbl')
        :param bool [params.trigger]: *swap only* whether to watch trigger orders(default is False)
        :returns dict[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        methodName = 'watchOrders'
        await self.load_markets()
        market = None
        marketId = None
        if symbol is not None:
            market = self.market(symbol)
            symbol = market['symbol']
            marketId = market['id']
        marketType = None
        marketType, params = self.handle_market_type_and_params(methodName, market, params)
        instType = 'spbl'
        instId = marketId
        if marketType == 'spot':
            if symbol is None:
                raise ArgumentsRequired(self.id + ' ' + methodName + '() requires a symbol argument for ' + marketType + ' markets.')
        else:
            instId = 'default'
            instType = 'umcbl'
            if symbol is None:
                instType, params = self.handle_option_and_params(params, methodName, 'instType', instType)
            else:
                if marketId.find('_DMCBL') >= 0:
                    instType = 'dmcbl'
        channel = 'orders'
        isTrigger = self.safe_bool(params, 'trigger')
        if isTrigger:
            channel = 'ordersAlgo'  # channel does not return any data
            params = self.omit(params, 'trigger')
        args: dict = {
            'instType': instType,
            'channel': channel,
            'instId': instId,
        }
        messageHash = 'orders'
        if symbol is not None:
            messageHash += ':' + symbol
        orders = await self.watch_private(messageHash, messageHash, args, params)
        if self.newUpdates:
            limit = orders.getLimit(symbol, limit)
        return self.filter_by_symbol_since_limit(orders, symbol, since, limit, True)

    def handle_order(self, client: Client, message):
        #
        # spot
        #
        #     {
        #         action: 'snapshot',
        #         arg: {instType: 'spbl', channel: 'orders', instId: 'ETHUSDT_SPBL'},
        #         data: [
        #             {
        #                 instId: 'ETHUSDT_SPBL',
        #                 ordId: '1228627925964996608',
        #                 clOrdId: 'f0cccf74-c535-4523-a53d-dbe3b9958559',
        #                 px: '2000',
        #                 sz: '0.001',
        #                 notional: '2',
        #                 ordType: 'limit',
        #                 force: 'normal',
        #                 side: 'buy',
        #                 accFillSz: '0',
        #                 avgPx: '0',
        #                 status: 'new',
        #                 cTime: 1728653645030,
        #                 uTime: 1728653645030,
        #                 orderFee: [],
        #                 eps: 'API'
        #             }
        #         ],
        #         ts: 1728653645046
        #     }
        #
        # swap
        #
        #     {
        #         action: 'snapshot',
        #         arg: {instType: 'umcbl', channel: 'orders', instId: 'default'},
        #         data: [
        #             {
        #                 accFillSz: '0',
        #                 cTime: 1728653796976,
        #                 clOrdId: '1228628563272753152',
        #                 eps: 'API',
        #                 force: 'normal',
        #                 hM: 'single_hold',
        #                 instId: 'ETHUSDT_UMCBL',
        #                 lever: '5',
        #                 low: False,
        #                 notionalUsd: '20',
        #                 ordId: '1228628563188867072',
        #                 ordType: 'limit',
        #                 orderFee: [],
        #                 posSide: 'net',
        #                 px: '2000',
        #                 side: 'buy',
        #                 status: 'new',
        #                 sz: '0.01',
        #                 tS: 'buy_single',
        #                 tdMode: 'cross',
        #                 tgtCcy: 'USDT',
        #                 uTime: 1728653796976
        #             }
        #         ],
        #         ts: 1728653797002
        #     }
        #
        #
        arg = self.safe_dict(message, 'arg', {})
        instType = self.safe_string(arg, 'instType')
        argInstId = self.safe_string(arg, 'instId')
        marketType = None
        if instType == 'spbl':
            marketType = 'spot'
        else:
            marketType = 'swap'
        data = self.safe_list(message, 'data', [])
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        hash = 'orders'
        stored = self.orders
        symbol: Str = None
        for i in range(0, len(data)):
            order = data[i]
            marketId = self.safe_string(order, 'instId', argInstId)
            market = self.safe_market(marketId, None, None, marketType)
            parsed = self.parse_ws_order(order, market)
            stored.append(parsed)
            symbol = parsed['symbol']
            messageHash = 'orders:' + symbol
            client.resolve(stored, messageHash)
        client.resolve(stored, hash)

    def parse_ws_order(self, order: dict, market: Market = None) -> Order:
        #
        # spot
        #     {
        #         instId: 'ETHUSDT_SPBL',
        #         ordId: '1228627925964996608',
        #         clOrdId: 'f0cccf74-c535-4523-a53d-dbe3b9958559',
        #         px: '2000',
        #         sz: '0.001',
        #         notional: '2',
        #         ordType: 'limit',
        #         force: 'normal',
        #         side: 'buy',
        #         accFillSz: '0',
        #         avgPx: '0',
        #         status: 'new',
        #         cTime: 1728653645030,
        #         uTime: 1728653645030,
        #         orderFee: orderFee: [{fee: '0', feeCcy: 'USDT'}],
        #         eps: 'API'
        #     }
        #
        # swap
        #     {
        #         accFillSz: '0',
        #         cTime: 1728653796976,
        #         clOrdId: '1228628563272753152',
        #         eps: 'API',
        #         force: 'normal',
        #         hM: 'single_hold',
        #         instId: 'ETHUSDT_UMCBL',
        #         lever: '5',
        #         low: False,
        #         notionalUsd: '20',
        #         ordId: '1228628563188867072',
        #         ordType: 'limit',
        #         orderFee: [{fee: '0', feeCcy: 'USDT'}],
        #         posSide: 'net',
        #         px: '2000',
        #         side: 'buy',
        #         status: 'new',
        #         sz: '0.01',
        #         tS: 'buy_single',
        #         tdMode: 'cross',
        #         tgtCcy: 'USDT',
        #         uTime: 1728653796976
        #     }
        #
        marketId = self.safe_string(order, 'instId')
        settleId = self.safe_string(order, 'tgtCcy')
        market = self.safeMarketCustom(marketId, market, settleId)
        timestamp = self.safe_integer(order, 'cTime')
        symbol = market['symbol']
        rawStatus = self.safe_string(order, 'status')
        orderFee = self.safe_list(order, 'orderFee', [])
        fee = self.safe_dict(orderFee, 0)
        feeCost = Precise.string_mul(self.safe_string(fee, 'fee'), '-1')
        feeCurrency = self.safe_string(fee, 'feeCcy')
        price = self.omit_zero(self.safe_string(order, 'px'))
        priceAvg = self.omit_zero(self.safe_string(order, 'avgPx'))
        if price is None:
            price = priceAvg
        type = self.safe_string_lower(order, 'ordType')
        return self.safe_order({
            'id': self.safe_string(order, 'ordId'),
            'clientOrderId': self.safe_string(order, 'clOrdId'),
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'lastUpdateTimestamp': self.safe_integer(order, 'uTime'),
            'status': self.parse_order_status(rawStatus),
            'symbol': symbol,
            'type': type,
            'timeInForce': self.parseOrderTimeInForce(self.safe_string_lower(order, 'force')),
            'side': self.safe_string_lower(order, 'side'),
            'price': price,
            'average': self.safe_string(order, 'avgPx'),
            'amount': self.safe_string(order, 'sz'),
            'filled': self.safe_string(order, 'accFillSz'),
            'remaining': None,
            'triggerPrice': None,
            'takeProfitPrice': None,
            'stopLossPrice': None,
            'cost': self.safe_string(order, 'notional'),
            'trades': None,
            'fee': {
                'currency': feeCurrency,
                'cost': feeCost,
            },
            'reduceOnly': self.safe_bool(order, 'low'),
            'postOnly': None,
            'info': order,
        }, market)

    async def watch_positions(self, symbols: Strings = None, since: Int = None, limit: Int = None, params={}) -> List[Position]:
        """
        watch all open positions

        https://coincatch.github.io/github.io/en/mix/#positions-channel

        :param str[]|None symbols: list of unified market symbols
 @param since
 @param limit
        :param dict params: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `position structure <https://docs.ccxt.com/en/latest/manual.html#position-structure>`
        """
        await self.load_markets()
        symbols = self.market_symbols(symbols, 'swap')
        messageHashes = []
        hash = 'positions'
        instTypes = []
        if symbols is not None:
            for i in range(0, len(symbols)):
                symbol = symbols[i]
                market = self.market(symbol)
                instType = self.get_private_inst_type(market)
                if not self.in_array(instType, instTypes):
                    instTypes.append(instType)
                messageHashes.append(hash + '::' + symbol)
        else:
            instTypes = ['umcbl', 'dmcbl']
            messageHashes.append(hash)
        args = []
        subscribeHashes = []
        for i in range(0, len(instTypes)):
            instType = instTypes[i]
            arg: dict = {
                'instType': instType,
                'channel': hash,
                'instId': 'default',
            }
            subscribeHashes.append(hash + '::' + instType)
            args.append(arg)
        newPositions = await self.watch_private_multiple(messageHashes, subscribeHashes, args, params)
        if self.newUpdates:
            return newPositions
        return self.filter_by_symbols_since_limit(newPositions, symbols, since, limit, True)

    def get_private_inst_type(self, market: Market):
        marketId = market['id']
        if marketId.find('_DMCBL') >= 0:
            return 'dmcbl'
        return 'umcbl'

    def handle_positions(self, client: Client, message):
        #
        #     {
        #         action: 'snapshot',
        #         arg: {instType: 'umcbl', channel: 'positions', instId: 'default'},
        #         data: [
        #             {
        #                 posId: '1221355728745619456',
        #                 instId: 'ETHUSDT_UMCBL',
        #                 instName: 'ETHUSDT',
        #                 marginCoin: 'USDT',
        #                 margin: '5.27182',
        #                 marginMode: 'crossed',
        #                 holdSide: 'long',
        #                 holdMode: 'single_hold',
        #                 total: '0.01',
        #                 available: '0.01',
        #                 locked: '0',
        #                 averageOpenPrice: '2635.91',
        #                 leverage: 5,
        #                 achievedProfits: '0',
        #                 upl: '-0.0267',
        #                 uplRate: '-0.005064664576',
        #                 liqPx: '-3110.66866033',
        #                 keepMarginRate: '0.0033',
        #                 marginRate: '0.002460827254',
        #                 cTime: '1726919818102',
        #                 uTime: '1728919604312',
        #                 markPrice: '2633.24',
        #                 autoMargin: 'off'
        #             }
        #         ],
        #         ts: 1728919604329
        #     }
        #
        if self.positions is None:
            self.positions = ArrayCacheBySymbolBySide()
        cache = self.positions
        rawPositions = self.safe_value(message, 'data', [])
        dataLength = len(rawPositions)
        if dataLength == 0:
            return
        newPositions = []
        symbols = []
        for i in range(0, len(rawPositions)):
            rawPosition = rawPositions[i]
            position = self.parse_ws_position(rawPosition)
            symbols.append(position['symbol'])
            newPositions.append(position)
            cache.append(position)
        hash = 'positions'
        messageHashes = self.find_message_hashes(client, hash)
        for i in range(0, len(messageHashes)):
            messageHash = messageHashes[i]
            parts = messageHash.split('::')
            symbol = parts[1]
            if self.in_array(symbol, symbols):
                positionsForSymbol = []
                for j in range(0, len(newPositions)):
                    position = newPositions[j]
                    if position['symbol'] == symbol:
                        positionsForSymbol.append(position)
                client.resolve(positionsForSymbol, messageHash)
        client.resolve(newPositions, hash)

    def parse_ws_position(self, position, market=None):
        #
        #     {
        #         posId: '1221355728745619456',
        #         instId: 'ETHUSDT_UMCBL',
        #         instName: 'ETHUSDT',
        #         marginCoin: 'USDT',
        #         margin: '5.27182',
        #         marginMode: 'crossed',
        #         holdSide: 'long',
        #         holdMode: 'single_hold',
        #         total: '0.01',
        #         available: '0.01',
        #         locked: '0',
        #         averageOpenPrice: '2635.91',
        #         leverage: 5,
        #         achievedProfits: '0',
        #         upl: '-0.0267',
        #         uplRate: '-0.005064664576',
        #         liqPx: '-3110.66866033',
        #         keepMarginRate: '0.0033',
        #         marginRate: '0.002460827254',
        #         cTime: '1726919818102',
        #         uTime: '1728919604312',
        #         markPrice: '2633.24',
        #         autoMargin: 'off'
        #     }
        #
        marketId = self.safe_string(position, 'symbol')
        settleId = self.safe_string(position, 'marginCoin')
        market = self.safeMarketCustom(marketId, market, settleId)
        timestamp = self.safe_integer(position, 'cTime')
        marginModeId = self.safe_string(position, 'marginMode')
        marginMode = self.get_supported_mapping(marginModeId, {
            'crossed': 'cross',
            'isolated': 'isolated',
        })
        isHedged: Bool = None
        holdMode = self.safe_string(position, 'holdMode')
        if holdMode == 'double_hold':
            isHedged = True
        elif holdMode == 'single_hold':
            isHedged = False
        percentageDecimal = self.safe_string(position, 'uplRate')
        percentage = Precise.string_mul(percentageDecimal, '100')
        margin = self.safe_number(position, 'margin')
        return self.safe_position({
            'symbol': market['symbol'],
            'id': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'contracts': self.safe_number(position, 'total'),
            'contractSize': None,
            'side': self.safe_string_lower(position, 'holdSide'),
            'notional': margin,  # todo check
            'leverage': self.safe_integer(position, 'leverage'),
            'unrealizedPnl': self.safe_number(position, 'upl'),
            'realizedPnl': self.safe_number(position, 'achievedProfits'),
            'collateral': None,  # todo check
            'entryPrice': self.safe_number(position, 'averageOpenPrice'),
            'markPrice': self.safe_number(position, 'markPrice'),
            'liquidationPrice': self.safe_number(position, 'liqPx'),
            'marginMode': marginMode,
            'hedged': isHedged,
            'maintenanceMargin': None,  # todo check
            'maintenanceMarginPercentage': self.safe_number(position, 'keepMarginRate'),
            'initialMargin': margin,  # todo check
            'initialMarginPercentage': None,
            'marginRatio': self.safe_number(position, 'marginRate'),
            'lastUpdateTimestamp': self.safe_integer(position, 'uTime'),
            'lastPrice': None,
            'stopLossPrice': None,
            'takeProfitPrice': None,
            'percentage': percentage,
            'info': position,
        })

    def handle_error_message(self, client: Client, message) -> Bool:
        #
        #    {event: "error", code: 30001, msg: "Channel does not exist"}
        #
        event = self.safe_string(message, 'event')
        try:
            if event == 'error':
                code = self.safe_string(message, 'code')
                feedback = self.id + ' ' + self.json(message)
                self.throw_exactly_matched_exception(self.exceptions['ws']['exact'], code, feedback)
                msg = self.safe_string(message, 'msg', '')
                self.throw_broadly_matched_exception(self.exceptions['ws']['broad'], msg, feedback)
                raise ExchangeError(feedback)
            return False
        except Exception as e:
            if isinstance(e, AuthenticationError):
                messageHash = 'authenticated'
                client.reject(e, messageHash)
                if messageHash in client.subscriptions:
                    del client.subscriptions[messageHash]
            else:
                client.reject(e)
            return True

    def handle_message(self, client: Client, message):
        # todo handle with subscribe and unsubscribe
        if self.handle_error_message(client, message):
            return
        content = self.safe_string(message, 'message')
        if content == 'pong':
            self.handle_pong(client, message)
            return
        if message == 'pong':
            self.handle_pong(client, message)
            return
        event = self.safe_string(message, 'event')
        if event == 'login':
            self.handle_authenticate(client, message)
            return
        if event == 'subscribe':
            self.handle_subscription_status(client, message)
            return
        if event == 'unsubscribe':
            self.handle_un_subscription_status(client, message)
            return
        data = self.safe_dict(message, 'arg', {})
        channel = self.safe_string(data, 'channel')
        if channel == 'ticker':
            self.handle_ticker(client, message)
        if channel.find('candle') >= 0:
            self.handle_ohlcv(client, message)
        if channel.find('books') >= 0:
            self.handle_order_book(client, message)
        if channel == 'trade':
            self.handle_trades(client, message)
        if channel == 'account':
            self.handle_balance(client, message)
        if (channel == 'orders') or (channel == 'ordersAlgo'):
            self.handle_order(client, message)
        if channel == 'positions':
            self.handle_positions(client, message)

    def ping(self, client: Client):
        return 'ping'

    def handle_pong(self, client: Client, message):
        client.lastPong = self.milliseconds()
        return message

    def handle_subscription_status(self, client: Client, message):
        return message

    def handle_un_subscription_status(self, client: Client, message):
        argsList = self.safe_list(message, 'args')
        if argsList is None:
            argsList = [self.safe_dict(message, 'arg', {})]
        for i in range(0, len(argsList)):
            arg = argsList[i]
            channel = self.safe_string(arg, 'channel')
            if channel == 'books':
                self.handle_order_book_un_subscription(client, message)
            elif channel == 'trade':
                self.handle_trades_un_subscription(client, message)
            elif channel == 'ticker':
                self.handle_ticker_un_subscription(client, message)
            elif channel.startswith('candle'):
                self.handle_ohlcv_un_subscription(client, message)
        return message

    def handle_order_book_un_subscription(self, client: Client, message):
        arg = self.safe_dict(message, 'arg', {})
        instType = self.safe_string_lower(arg, 'instType')
        type = 'spot' if (instType == 'sp') else 'swap'
        instId = self.safe_string(arg, 'instId')
        market = self.safe_market(instId, None, None, type)
        symbol = market['symbol']
        messageHash = 'unsubscribe:orderbook:' + market['symbol']
        subMessageHash = 'orderbook:' + symbol
        if symbol in self.orderbooks:
            del self.orderbooks[symbol]
        if subMessageHash in client.subscriptions:
            del client.subscriptions[subMessageHash]
        if messageHash in client.subscriptions:
            del client.subscriptions[messageHash]
        error = UnsubscribeError(self.id + ' orderbook ' + symbol)
        client.reject(error, subMessageHash)
        client.resolve(True, messageHash)

    def handle_trades_un_subscription(self, client: Client, message):
        arg = self.safe_dict(message, 'arg', {})
        instType = self.safe_string_lower(arg, 'instType')
        type = 'spot' if (instType == 'sp') else 'swap'
        instId = self.safe_string(arg, 'instId')
        market = self.safe_market(instId, None, None, type)
        symbol = market['symbol']
        messageHash = 'unsubscribe:trade:' + market['symbol']
        subMessageHash = 'trade:' + symbol
        if symbol in self.trades:
            del self.trades[symbol]
        if subMessageHash in client.subscriptions:
            del client.subscriptions[subMessageHash]
        if messageHash in client.subscriptions:
            del client.subscriptions[messageHash]
        error = UnsubscribeError(self.id + ' trades ' + symbol)
        client.reject(error, subMessageHash)
        client.resolve(True, messageHash)

    def handle_ticker_un_subscription(self, client: Client, message):
        arg = self.safe_dict(message, 'arg', {})
        instType = self.safe_string_lower(arg, 'instType')
        type = 'spot' if (instType == 'sp') else 'swap'
        instId = self.safe_string(arg, 'instId')
        market = self.safe_market(instId, None, None, type)
        symbol = market['symbol']
        messageHash = 'unsubscribe:ticker:' + market['symbol']
        subMessageHash = 'ticker:' + symbol
        if symbol in self.tickers:
            del self.tickers[symbol]
        if subMessageHash in client.subscriptions:
            del client.subscriptions[subMessageHash]
        if messageHash in client.subscriptions:
            del client.subscriptions[messageHash]
        error = UnsubscribeError(self.id + ' ticker ' + symbol)
        client.reject(error, subMessageHash)
        client.resolve(True, messageHash)

    def handle_ohlcv_un_subscription(self, client: Client, message):
        arg = self.safe_dict(message, 'arg', {})
        instType = self.safe_string_lower(arg, 'instType')
        type = 'spot' if (instType == 'sp') else 'swap'
        instId = self.safe_string(arg, 'instId')
        channel = self.safe_string(arg, 'channel')
        interval = channel.replace('candle', '')
        timeframes = self.safe_dict(self.options, 'timeframesForWs')
        timeframe = self.find_timeframe(interval, timeframes)
        market = self.safe_market(instId, None, None, type)
        symbol = market['symbol']
        messageHash = 'unsubscribe:ohlcv:' + timeframe + ':' + market['symbol']
        subMessageHash = 'ohlcv:' + symbol + ':' + timeframe
        if symbol in self.ohlcvs:
            if timeframe in self.ohlcvs[symbol]:
                del self.ohlcvs[symbol][timeframe]
        self.clean_unsubscription(client, subMessageHash, messageHash)



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_api_spot_product_list = publicGetApiSpotProductList = Entry('api/spot/product/list', 'public', 'GET', {'cost': 1})
    public_get_api_tv_tradingview_history = publicGetApiTvTradingViewHistory = Entry('api/tv/tradingView/history', 'public', 'GET', {'cost': 1})
    quote_get_tickers = quoteGetTickers = Entry('tickers', 'quote', 'GET', {'cost': 1})
    quote_get_mkpai_depth_v2 = quoteGetMkpaiDepthV2 = Entry('mkpai/depth-v2', 'quote', 'GET', {'cost': 1})
